from scipy.stats import gaussian_kde
from bokeh.models import ColumnDataSource, LinearColorMapper, ColorBar, InlineStyleSheet
from bokeh.transform import linear_cmap
from bokeh.palettes import Viridis256
import bokeh.models
import bokeh.palettes
import re
import numpy as np


def color_by_feature_value(
    points_bokeh_plot,
    violins_bokeh_plot,
    heat_map,
    adata,
    select_color_by,
    hidden_text_label_column,
    resize_width_input,
    hidden_legend_width,
    hidden_checkbox_A,
    hidden_checkbox_B,
    resize_w_input,
    resize_h_input,
    bt_slider_range,
):
    source = points_bokeh_plot.select(dict(name="scatterplot"))[0].data_source

    feature_dict = {}
    feat_min = np.min(adata.X, axis=0)
    feat_max = np.max(adata.X, axis=0)

    for i, featname in enumerate(adata.var_names):
        feature_dict[featname] = [i, feat_min[i], feat_max[i]]

    def color_modif(
        stringval, htlc, rwi, hlw, label_stringval, resize_w, resize_h, bt_slider_range
    ):
        stringval_modif = ("  +  " + stringval).replace("  +    -  ", "  -  ").replace(
            "  +    +  ", "  +  "
        ).replace("  +  ", "§§§§§§§§§§  +  ").replace(
            "  -  ", "§§§§§§§§§§  -  "
        ) + "§§§§§§§§§§"
        positive_matches = [
            elt[5:-10] for elt in re.findall("  \\+  .*?§§§§§§§§§§", stringval_modif)
        ]
        negative_matches = [
            elt[5:-10] for elt in re.findall("  \\-  .*?§§§§§§§§§§", stringval_modif)
        ]

        if label_stringval:
            selected_labels = [
                lbl.strip() for lbl in label_stringval.split("//yomix//") if lbl.strip()
            ]
        else:
            selected_labels = None

        plot_var_features = []
        for elt in positive_matches + negative_matches:
            if elt in feature_dict:
                plot_var_features += [elt]

        if len(plot_var_features) > 0 and selected_labels is not None:
            plot_var(
                adata,
                points_bokeh_plot,
                violins_bokeh_plot,
                heat_map,
                resize_w,
                resize_h,
                hidden_checkbox_A,
                hidden_checkbox_B,
                features=plot_var_features,
                selected_labels=selected_labels,
            )

        if len(positive_matches) + len(negative_matches) > 0:
            contn = True
            for elt in positive_matches + negative_matches:
                if elt not in feature_dict:
                    contn = False
            if contn:
                if len(positive_matches) == 1 and len(negative_matches) == 0:
                    elt = positive_matches[0]
                    vmin = feature_dict[elt][1]
                    vmax = feature_dict[elt][2]
                    new_data_color = (
                        adata.X[source.data["index"][:], feature_dict[elt][0]] - vmin
                    ) / (vmax - vmin + 0.000001)
                elif len(positive_matches) == 0 and len(negative_matches) == 1:
                    elt = negative_matches[0]
                    vmax = -feature_dict[elt][1]
                    vmin = -feature_dict[elt][2]
                    new_data_color = (
                        -adata.X[source.data["index"][:], feature_dict[elt][0]] - vmin
                    ) / (vmax - vmin + 0.000001)
                else:
                    new_data_color = np.zeros(len(source.data["color"]))
                    for elt in positive_matches:
                        vmin = feature_dict[elt][1]
                        vmax = feature_dict[elt][2]
                        new_data_color += (
                            adata.X[source.data["index"][:], feature_dict[elt][0]]
                            - vmin
                        ) / (vmax - vmin + 0.000001)
                    for elt in negative_matches:
                        vmax = -feature_dict[elt][1]
                        vmin = -feature_dict[elt][2]
                        new_data_color += (
                            -adata.X[source.data["index"][:], feature_dict[elt][0]]
                            - vmin
                        ) / (vmax - vmin + 0.000001)
                    new_data_color = new_data_color / (new_data_color.max() + 0.000001)
                source.data["color_ref"] = new_data_color

                points_bokeh_plot.right = []
                htlc.value = ""
                viridis_colors = list(bokeh.palettes.Viridis256)
                custom_color_mapper = bokeh.models.LinearColorMapper(
                    palette=viridis_colors, low=0.0, high=1.0
                )

                def simple_shrink(s_in, size):
                    true_size = max(size, 3)
                    if len(s_in) > true_size:
                        new_s = ""
                        l1 = true_size - 3
                        new_s += s_in[:l1]
                        new_s += "..."
                    else:
                        new_s = s_in
                    return new_s

                cbar = bokeh.models.ColorBar(
                    color_mapper=custom_color_mapper,
                    label_standoff=12,
                    width=47,
                    ticker=bokeh.models.FixedTicker(ticks=[]),
                    title=simple_shrink(stringval, 50),
                )
                points_bokeh_plot.add_layout(cbar, "right")
                # label_font_size = cbar.major_label_text_font_size
                # label_font_size = int(label_font_size[:-2])
                legend_width = 55
                rwi.value = str(
                    int(points_bokeh_plot.width - float(hlw.value) + legend_width)
                )
                hlw.value = str(int(legend_width))
                select_color_by.value = ""
                current_style = bt_slider_range.stylesheets[0].css
                pattern = r"\{margin: 32px 0px 0px -\d+px;\}"
                new_style = re.sub(
                    pattern,
                    "{margin: 32px 0px 0px -" + str(int(legend_width)) + "px;}",
                    current_style,
                )
                bt_slider_range.stylesheets = [InlineStyleSheet(css=new_style)]
                bt_slider_range.start = 0.0
                bt_slider_range.end = 1.0
                bt_slider_range.value = (0.0, 1.0)
                bt_slider_range.step = 0.01
                bt_slider_range.visible = True

    source.js_on_change(
        "data",
        bokeh.models.CustomJS(
            args=dict(source=source),
            code="""
        const data = source.data;
        for (let i = 0; i < data["color"].length; i++) {
            data["color"][i] = data["color_ref"][data["index"][i]];
        }
    """,
        ),
    )

    offset_text_feature_color = bokeh.models.TextInput(
        value="",
        title="Color by feature value (enter feature name):",
        name="offset_text_feature_color",
        width=650,
    )

    offset_label = bokeh.models.TextInput(
        value="",
        title="Selected Groups:",
        name=" _plot",
        width=650,
    )

    offset_text_feature_color.on_change(
        "value",
        lambda attr, old, new: color_modif(
            new,
            hidden_text_label_column,
            resize_width_input,
            hidden_legend_width,
            offset_label.value,  # Include the current label value
            resize_w_input,
            resize_h_input,
            bt_slider_range,
        ),
    )

    # Modify the callback for offset_label
    offset_label.on_change(
        "value",
        lambda attr, old, new_label: color_modif(
            offset_text_feature_color.value,
            hidden_text_label_column,
            resize_width_input,
            hidden_legend_width,
            new_label,  # Pass the new label value to color_modif
            resize_w_input,
            resize_h_input,
            bt_slider_range,
        ),
    )

    return offset_text_feature_color, offset_label


def plot_var(
    adata,
    points_bokeh_plot,
    violins_bokeh_plot,
    heat_map,
    resize_w,
    resize_h,
    hidden_checkbox_A,
    hidden_checkbox_B,
    features,
    selected_labels=None,  # This is where the selected labels will be passed
    equal_size=False,
):

    def get_kde(data, grid_points=100):
        kde = gaussian_kde(data)
        y = np.linspace(np.min(data), np.max(data), grid_points)
        x = kde(y)
        return x, y

    def plot_violin_from_feat(xd, feat, labels, mode="violin"):
        data_tmp = {"x": [], "y": [], "median_expr": [], "alpha": []}
        step = 0
        labels_nr = len(labels)
        for label in labels:
            if label == "[  Subset A  ]":
                data = xd[hidden_checkbox_A.active, feat].X.toarray().reshape(-1)
                if np.any(data):
                    median_expr = np.median(data)
            elif label == "[  Subset B  ]":
                data = xd[hidden_checkbox_B.active, feat].X.toarray().reshape(-1)
                if np.any(data):
                    median_expr = np.median(data)
            elif label == "[  Rest  ]":
                idx = np.arange(adata.n_obs)[
                    ~np.isin(np.arange(xd.n_obs), hidden_checkbox_A.active)
                ]
                data = xd[idx, feat].X.toarray().reshape(-1)
                if np.any(data):
                    median_expr = np.median(data)
            else:
                lblsplit = label.split(">>yomix>>")
                lbl = lblsplit[0]
                lbl_elt = lblsplit[1]
                data = xd[xd.obs[lbl] == lbl_elt, feat].X.toarray().reshape(-1)
                if np.any(data):
                    if xd.var["yomix_median_" + label][feat] < 0:
                        xd.var.loc[feat, "yomix_median_" + label] = np.median(data)
                    median_expr = xd.var["yomix_median_" + label][feat]
            if np.any(data) and (len(data) > 1 or mode != "violin"):
                if mode == "violin":
                    x, y = get_kde(data)
                else:  # heatmap
                    x, y = np.ones(100), np.linspace(0, 1, 100)
                # same width for every subset
                x = (2.5 - np.clip(0.01 * labels_nr, 0, 0.1)) * x / np.max(x)
                data_tmp["x"].append(np.concatenate([x, -x[::-1]]) + step)
                data_tmp["y"].append(np.concatenate([y, y[::-1]]))
                data_tmp["median_expr"].append(median_expr)
                data_tmp["alpha"].append(1.0)
            elif np.any(data) and len(data) == 1 and mode == "violin":
                bound = 2.5 - np.clip(0.01 * labels_nr, 0, 0.1)
                line = np.linspace(step - bound, step + bound, 100)
                data_tmp["x"].append(line)
                data_tmp["y"].append([median_expr for i in line])
                data_tmp["median_expr"].append(median_expr)
                data_tmp["alpha"].append(1.0)
            else:
                if mode == "violin":
                    # line = np.linspace(step - 1, step + 1, 100)
                    bound = 2.5 - np.clip(0.01 * labels_nr, 0, 0.1)
                    line = np.linspace(step - bound, step + bound, 100)
                    data_tmp["x"].append(line)
                    data_tmp["y"].append([0 for i in line])
                    data_tmp["median_expr"].append(0)
                    data_tmp["alpha"].append(1.0)
                else:
                    x, y = np.ones(100), np.linspace(0, 1, 100)
                    x = (2.5 - np.clip(0.01 * labels_nr, 0, 0.1)) * x / np.max(x)
                    data_tmp["x"].append(np.concatenate([x, -x[::-1]]) + step)
                    data_tmp["y"].append(np.concatenate([y, y[::-1]]))
                    data_tmp["median_expr"].append(0)
                    data_tmp["alpha"].append(1.0)
            step += 5
        bound = 2.5 - np.clip(0.01 * labels_nr, 0, 0.1)
        data_tmp["x"].append(np.linspace(0 - bound, step - 5 + bound, 100))
        data_tmp["y"].append(np.ones(100))
        data_tmp["median_expr"].append(0)
        data_tmp["alpha"].append(0.2)
        data_tmp["x"].append(np.linspace(0 - bound, step - 5 + bound, 100))
        data_tmp["y"].append(0.0 * np.ones(100))
        data_tmp["median_expr"].append(0)
        data_tmp["alpha"].append(0.2)
        return data_tmp

    def refresh_violin(vplot, mode="violin"):
        data_tmp = {"x": [], "y": [], "median_expr": [], "alpha": []}

        step_yaxis = 0
        if selected_labels is None:
            labels = None
        else:
            labels = selected_labels

        for feat in features[::-1]:
            tmp_dict = plot_violin_from_feat(adata, feat, labels, mode=mode)
            tmp_dict["y"] = [np.asarray(i) + step_yaxis for i in tmp_dict["y"]]
            for key in data_tmp.keys():
                data_tmp[key].extend(tmp_dict[key])
            step_yaxis += 1.1

        vplot.yaxis.major_label_text_font_size = "10pt"
        if len(features) > 1:
            set_yticks = [0.5 + 1.1 * i for i in range(len(features))]
            vplot.yaxis.ticker = set_yticks
            vplot.yaxis.major_label_overrides = {
                set_yticks[i]: features[len(features) - i - 1]
                for i in range(len(features))
            }
            vplot.yaxis.axis_label = ""
        else:
            vplot.yaxis.axis_label = features[0]
            set_yticks = []
            vplot.yaxis.ticker = set_yticks
            vplot.yaxis.major_label_overrides = {}

        custom_color_mapper = LinearColorMapper(palette=Viridis256, low=0, high=1)

        # check if patches already exist
        if vplot.select(name="violins"):
            source = vplot.select(dict(name="violins"))[0].data_source
            source.data = data_tmp
        else:
            source = ColumnDataSource(data=data_tmp)
            vplot.patches(
                "x",
                "y",
                source=source,
                fill_color=linear_cmap(
                    "median_expr", palette=Viridis256, low=0, high=1
                ),
                line_alpha="alpha",
                line_color="black",
                name="violins",
            )
            color_bar = ColorBar(
                color_mapper=custom_color_mapper,
                title="Normalized median feature value in group",
                title_text_align="center",
                major_tick_line_color=None,
                major_label_text_font_size="0pt",
            )
            vplot.add_layout(color_bar, "right")

        vplot.title.text = "Feature values per group"
        vplot.xgrid.visible = False
        vplot.ygrid.visible = False

        set_xticks = list(range(0, len(labels) * 5, 5))
        vplot.xaxis.ticker = set_xticks
        samples_per_labels = {}
        text_labels = []
        for label in labels:
            if label == "[  Subset A  ]":
                samples_per_labels["[  Subset A  ]"] = str(
                    len(hidden_checkbox_A.active)
                )
                text_labels.append(label + "\n ")
            elif label == "[  Subset B  ]":
                samples_per_labels["[  Subset B  ]"] = str(
                    len(hidden_checkbox_B.active)
                )
                text_labels.append(label + "\n ")
            elif label == "[  Rest  ]":
                samples_per_labels["[  Rest  ]"] = str(
                    len(adata) - len(hidden_checkbox_A.active)
                )
                text_labels.append(label + "\n ")
            else:
                lblsplit = label.split(">>yomix>>")
                lbl = lblsplit[0]
                lbl_elt = lblsplit[1]
                samples_per_labels[label] = str(len(adata[adata.obs[lbl] == lbl_elt]))
                text_labels.append(lbl + "\n" + lbl_elt)

        vplot.xaxis.major_label_overrides = {
            set_xticks[i]: (
                text_labels[i] + "\n\n" + samples_per_labels[labels[i]] + "\nsamples"
                if int(samples_per_labels[labels[i]]) > 1
                else text_labels[i]
                + "\n\n"
                + samples_per_labels[labels[i]]
                + "\nsample"
            )
            for i in range(len(set_xticks))
        }

        vplot.visible = True

    refresh_violin(violins_bokeh_plot, mode="violin")
    refresh_violin(heat_map, mode="heatmap")

    resize_w.visible = True
    resize_h.visible = True



""" Enable execution of the "yomix" command line program with the ``-m``
switch. For example:

.. code-block:: sh

    python -m yomix myfile.h5ad

is equivalent to

.. code-block:: sh

    yomix myfile.h5ad

"""

import yomix
import numpy as np
import bokeh.layouts
import anndata
from scipy.sparse import issparse
from tornado.ioloop import IOLoop
from bokeh.application.handlers import FunctionHandler
from bokeh.application import Application
from bokeh.server.server import Server
from bokeh.models import TabPanel, Tabs
from pathlib import Path
import argparse

__all__ = ("main",)


def main():

    parser = argparse.ArgumentParser(description="Yomix command-line tool")

    parser.add_argument(
        "file", type=str, nargs="?", default=None, help="the .ha5d file to open"
    )
    parser.add_argument(
        "--subsampling",
        type=int,
        help="randomly subsample the dataset to a maximum number of observations "
        "(=SUBSAMPLING)",
    )
    parser.add_argument(
        "--example", action="store_true", help="use the example dataset"
    )

    args = parser.parse_args()

    argument = args.example

    if argument:
        filearg = Path(__file__).parent / "example" / "pbmc.h5ad"
    else:
        assert (
            args.file is not None
        ), "yomix: error: the following arguments are required: file"
        filearg = Path(args.file)

    if filearg.exists():
        xd = anndata.read_h5ad(filearg.absolute())

        def _to_dense(x):
            if issparse(x):
                return x.todense()
            else:
                return x

        xd.X = np.asarray(_to_dense(xd.X))

        if args.subsampling is not None:
            if xd.n_obs > args.subsampling:
                selected_obs = np.random.choice(
                    xd.n_obs, args.subsampling, replace=False
                )
                selected_obs.sort()
                xd = xd[selected_obs].copy()

        min_norm = np.min(xd.X, axis=0)
        max_norm = np.max(xd.X, axis=0)
        xd.X = np.divide(xd.X - min_norm, max_norm - min_norm + 1e-8)
        obs_string_init = list(xd.obs.select_dtypes("category").keys())
        all_labels_list = []
        for lbl in obs_string_init:
            # filter labels
            if yomix.plotting.check_obs_field(xd, str(lbl)):
                labels = np.array(list(dict.fromkeys(xd.obs[str(lbl)])))
                all_labels_list += [(str(lbl), str(elt)) for elt in labels]
                for elt in labels:
                    xd.var["yomix_median_" + str(lbl) + ">>yomix>>" + str(elt)] = (
                        -np.ones(xd.n_vars)
                    )
        xd.uns["all_labels"] = all_labels_list

        def var_mean_values(adata) -> np.ndarray:
            return np.squeeze(np.asarray(np.mean(adata.X, axis=0)))

        def var_standard_deviations(adata) -> np.ndarray:
            return np.squeeze(np.asarray(np.std(adata.X, axis=0)))

        xd.var["mean_values_local_yomix"] = var_mean_values(xd)
        xd.var["standard_deviations_local_yomix"] = var_standard_deviations(xd)
        xd.var_names_make_unique()
        xd.obs_names_make_unique()

        def modify_doc(doc):

            def build_figure(embedding_key):

                if embedding_key is None:
                    embedding_key = ""

                list_ok_embed_keys = []
                for k in xd.obsm.keys():
                    if xd.obsm[k].shape[1] > 1:
                        list_ok_embed_keys.append(k)

                bt_select_embedding = bokeh.models.Select(
                    title="Select embedding (.obsm field)",
                    value=embedding_key,
                    width=235,
                    options=[(k, k) for k in list_ok_embed_keys],
                    name="bt_select_embedding",
                )

                if embedding_key != "":

                    embedding_size = xd.obsm[embedding_key].shape[1]
                    assert embedding_size > 1

                    (
                        obs_string,
                        obs_numerical,
                        points_bokeh_plot,
                        violins_bokeh_plot,
                        heat_map,
                        bt_slider_point_size,
                        bt_hidden_slider_yaw,
                        bt_slider_range,
                        bt_toggle_anim,
                        bt_slider_yaw,
                        bt_slider_pitch,
                        bt_slider_roll,
                        resize_width_input,
                        resize_height_input,
                        resize_width_input_bis,
                        resize_height_input_bis,
                        source_rotmatrix_etc,
                        div_sample_names,
                        sample_search_input,
                        sl_component1,
                        sl_component2,
                        sl_component3,
                    ) = yomix.plotting.main_figure(xd, embedding_key, 890, 390, "")

                    resize_width_input_bis.visible = False
                    resize_height_input_bis.visible = False

                    (
                        bt_A,
                        toggle_A,
                        hidden_checkbox_A,
                        bt_B,
                        toggle_B,
                        hidden_checkbox_B,
                        bt_AplusB,
                        bt_nothing,
                    ) = yomix.tools.subset_buttons(
                        points_bokeh_plot, source_rotmatrix_etc
                    )

                    (
                        select_color_by,
                        help_color_by,
                        hidden_text_label_column,
                        hidden_legend_width,
                    ) = yomix.plotting.setup_legend(
                        points_bokeh_plot,
                        obs_string,
                        obs_numerical,
                        source_rotmatrix_etc,
                        resize_width_input,
                        bt_slider_range,
                    )

                    offset_text_feature_color, offset_label = (
                        yomix.plotting.color_by_feature_value(
                            points_bokeh_plot,
                            violins_bokeh_plot,
                            heat_map,
                            xd,
                            select_color_by,
                            hidden_text_label_column,
                            resize_width_input,
                            hidden_legend_width,
                            hidden_checkbox_A,
                            hidden_checkbox_B,
                            resize_width_input_bis,
                            resize_height_input_bis,
                            bt_slider_range,
                        )
                    )
                    offset_label.visible = False

                    (
                        bt_sign1,
                        bt_sign2,
                        help1,
                        help2,
                        multiselect_signature,
                        div_signature_list,
                        sign_nr,
                        label_signature,
                    ) = yomix.tools.signature_buttons(
                        xd,
                        offset_text_feature_color,
                        offset_label,
                        hidden_checkbox_A,
                        hidden_checkbox_B,
                    )

                    select_color_by.js_on_change(
                        "value",
                        bokeh.models.CustomJS(
                            args=dict(
                                otfc=offset_text_feature_color, ms=multiselect_signature
                            ),
                            code="""
                                if (cb_obj.value != "") {
                                    otfc.value="";
                                    ms.value=[];
                                }
                            """,
                        ),
                    )

                    bt_open_link = yomix.tools.gene_query_button(
                        offset_text_feature_color
                    )

                    bt_sign3, help3 = yomix.tools.arrow_function(
                        points_bokeh_plot,
                        xd,
                        embedding_key,
                        bt_slider_roll,
                        bt_slider_pitch,
                        bt_slider_yaw,
                        source_rotmatrix_etc,
                        bt_toggle_anim,
                        hidden_checkbox_A,
                        div_signature_list,
                        multiselect_signature,
                        sign_nr,
                        sl_component1,
                        sl_component2,
                        sl_component3,
                        label_signature,
                    )

                    c1div = bokeh.models.Div(text="X axis:")
                    c2div = bokeh.models.Div(text="Y axis:")
                    c3div = bokeh.models.Div(text="Z axis:")

                    if embedding_size == 2:
                        bt_slider_yaw.visible = False
                        bt_slider_pitch.visible = False
                        bt_toggle_anim.visible = False
                        bt_toggle_anim.active = False
                        c3div.visible = False

                    violins_bokeh_plot.toolbar.logo = None
                    violins_bokeh_plot.toolbar_location = None
                    heat_map.toolbar.logo = None
                    heat_map.toolbar_location = None

                    tabs = Tabs(
                        tabs=[
                            TabPanel(child=violins_bokeh_plot, title="Violin plots"),
                            TabPanel(child=heat_map, title="Heatmap"),
                        ]
                    )

                    p = bokeh.layouts.row(
                        bokeh.layouts.column(
                            bt_select_embedding,
                            bokeh.layouts.row(bt_A, toggle_A),
                            bokeh.layouts.row(bt_B, toggle_B),
                            bokeh.layouts.row(bt_nothing, bt_AplusB),
                            bokeh.layouts.row(bt_sign1, help1),
                            bokeh.layouts.row(bt_sign2, help2),
                            bokeh.layouts.row(bt_sign3, help3),
                            multiselect_signature,
                            label_signature,
                            div_signature_list,
                        ),
                        (
                            bokeh.layouts.column(
                                bokeh.layouts.row(
                                    bokeh.layouts.column(c1div, c2div, c3div),
                                    bokeh.layouts.column(
                                        sl_component1, sl_component2, sl_component3
                                    ),
                                ),
                                bokeh.layouts.row(
                                    bokeh.layouts.row(
                                        bokeh.layouts.column(
                                            bt_slider_roll,
                                            bt_slider_yaw,
                                            bt_toggle_anim,
                                        ),
                                        bt_slider_pitch,
                                        bt_slider_point_size,
                                    ),
                                    bokeh.layouts.column(
                                        bokeh.layouts.row(
                                            select_color_by,
                                            help_color_by,
                                            sample_search_input,
                                        ),
                                        bokeh.layouts.row(
                                            offset_text_feature_color, bt_open_link
                                        ),
                                    ),
                                ),
                                bokeh.layouts.column(
                                    bokeh.layouts.row(
                                        bokeh.layouts.column(
                                            resize_height_input, resize_width_input
                                        ),
                                        points_bokeh_plot,
                                        bt_slider_range,
                                    ),
                                    bokeh.layouts.row(
                                        bokeh.layouts.column(
                                            resize_height_input_bis,
                                            resize_width_input_bis,
                                        ),
                                        tabs,
                                        div_sample_names,
                                    ),
                                ),
                                offset_label,
                            )
                            if sl_component1 is not None
                            else bokeh.layouts.column(
                                bokeh.layouts.row(
                                    bokeh.layouts.row(
                                        bokeh.layouts.column(
                                            bt_slider_roll,
                                            bt_slider_yaw,
                                            bt_toggle_anim,
                                        ),
                                        bt_slider_pitch,
                                        bt_slider_point_size,
                                    ),
                                    bokeh.layouts.column(
                                        bokeh.layouts.row(
                                            select_color_by,
                                            help_color_by,
                                            sample_search_input,
                                        ),
                                        bokeh.layouts.row(
                                            offset_text_feature_color, bt_open_link
                                        ),
                                    ),
                                ),
                                bokeh.layouts.column(
                                    bokeh.layouts.row(
                                        bokeh.layouts.column(
                                            resize_height_input, resize_width_input
                                        ),
                                        points_bokeh_plot,
                                        bt_slider_range,
                                    ),
                                    bokeh.layouts.row(
                                        bokeh.layouts.column(
                                            resize_height_input_bis,
                                            resize_width_input_bis,
                                        ),
                                        tabs,
                                        div_sample_names,
                                    ),
                                ),
                                offset_label,
                            )
                        ),
                        bt_hidden_slider_yaw,
                    )

                else:
                    p = bokeh.layouts.row(bt_select_embedding)

                p.name = "root"
                return p

            def reset_figure(new):
                doc.clear()
                p_new = build_figure(new)
                p_new.select_one(dict(name="bt_select_embedding")).on_change(
                    "value", lambda attr, old, new: reset_figure(new)
                )
                doc.add_root(p_new)

            p_0 = build_figure(None)
            p_0.select_one(dict(name="bt_select_embedding")).on_change(
                "value", lambda attr, old, new: reset_figure(new)
            )

            doc.add_root(p_0)

            def f():
                slider = doc.get_model_by_name("root").select_one(
                    dict(name="bt_hidden_slider_yaw")
                )
                anim = doc.get_model_by_name("root").select_one(
                    dict(name="bt_toggle_anim")
                )
                # print(slider)
                if slider is not None and anim.active:
                    slider.value = 10

            doc.add_periodic_callback(f, 100)

        io_loop = IOLoop.current()

        bokeh_app = Application(FunctionHandler(modify_doc))

        server = Server({"/": bokeh_app}, io_loop=io_loop)
        server.start()

        print("Opening Yomix on http://localhost:5006/\n")

        io_loop.add_callback(server.show, "/")
        io_loop.start()



from pathlib import Path
import pandas as pd
import bokeh.models
import numpy as np


def arrow_function(
    points_bokeh_plot,
    adata,
    embedding_key,
    bt_slider_roll,
    bt_slider_pitch,
    bt_slider_yaw,
    source_rotmatrix_etc,
    bt_toggle_anim,
    hidden_checkbox_A,
    div_signature_list,
    multiselect_signature,
    sign_nr,
    sl_component1,
    sl_component2,
    sl_component3,
    label_sign,
):

    arrow_clicks = bokeh.models.ColumnDataSource(data=dict(x=[], y=[]))

    arrow = bokeh.models.Arrow(
        end=bokeh.models.OpenHead(line_color="darkgray", line_width=4),
        line_color="darkgray",
        line_width=4,
        x_start=0.0,
        x_end=0.0,
        y_start=0.0,
        y_end=0.0,
        name="arrow",
        visible=False,
    )

    points_bokeh_plot.add_layout(arrow)

    def callback_arrow(event):
        if arrow.visible:
            if len(arrow_clicks.data["x"]) == 0:
                arrow_clicks.stream(dict(x=[event.x], y=[event.y]))
            else:
                arrow.x_start, arrow.y_start = (
                    arrow_clicks.data["x"][0],
                    arrow_clicks.data["y"][0],
                )
                arrow.x_end, arrow.y_end = event.x, event.y
                arrow_clicks.data = dict(x=[], y=[])

    points_bokeh_plot.on_event(bokeh.events.Tap, callback_arrow)

    hidden_toggle = bokeh.models.Toggle(name="hidden_toggle", active=False)

    arrow_tool = bokeh.models.CustomAction(
        description="Arrow Tool: click once for the start, twice for the end" "",
        icon=(Path(__file__).parent.parent / "assets" / "arrow.png").absolute(),
        callback=bokeh.models.CustomJS(
            args=dict(arr=arrow, hidden_t=hidden_toggle, btta=bt_toggle_anim),
            code="""
        btta.active=false;
        hidden_t.active=!hidden_t.active;
    """,
        ),
    )
    points_bokeh_plot.add_tools(arrow_tool)

    def toggle_arrow(new):
        if new:
            arrow_tool.icon = (
                Path(__file__).parent.parent / "assets" / "arrow_pressed.png"
            ).absolute()
            delta_x = (
                points_bokeh_plot.x_range.end - points_bokeh_plot.x_range.start
            ) / 4.0
            delta_y = (
                points_bokeh_plot.y_range.end - points_bokeh_plot.y_range.start
            ) / 4.0
            arrow.x_start = points_bokeh_plot.x_range.start + delta_x
            arrow.y_start = points_bokeh_plot.y_range.start + delta_y
            arrow.x_end = points_bokeh_plot.x_range.end - delta_x
            arrow.y_end = points_bokeh_plot.y_range.end - delta_y
            arrow.visible = True
        else:
            arrow_tool.icon = (
                Path(__file__).parent.parent / "assets" / "arrow.png"
            ).absolute()
            arrow.visible = False

    hidden_toggle.on_change("active", lambda attr, old, new: toggle_arrow(new))

    def rotations_deactivate_arrow():
        if hidden_toggle.active:
            hidden_toggle.active = False

    def toggle_rotations_deactivate_arrow(new):
        if new:
            if hidden_toggle.active:
                hidden_toggle.active = False

    bt_slider_roll.on_change(
        "value", lambda attr, old, new: rotations_deactivate_arrow()
    )
    bt_slider_pitch.on_change(
        "value", lambda attr, old, new: rotations_deactivate_arrow()
    )
    bt_slider_yaw.on_change(
        "value", lambda attr, old, new: rotations_deactivate_arrow()
    )
    bt_toggle_anim.on_change(
        "active", lambda attr, old, new: toggle_rotations_deactivate_arrow(new)
    )

    hidden_numeric_inputs = [
        bokeh.models.NumericInput(mode="float", value=0.0) for _ in range(10)
    ]

    tooltip = bokeh.models.Tooltip(
        content="Requires drawing an arrow with the Arrow Tool "
        "and setting subset A.\u00A0\u00A0",
        position="right",
    )
    help_button_oriented = bokeh.models.HelpButton(tooltip=tooltip, margin=(3, 0, 3, 0))
    bt_sign_oriented = bokeh.models.Button(
        label="Compute oriented signature (A)", width=190, margin=(5, 0, 5, 5)
    )

    bt_sign_oriented.js_on_click(
        bokeh.models.CustomJS(
            args=dict(
                source_rotmatrix_etc=source_rotmatrix_etc,
                numinputs=hidden_numeric_inputs,
                ht=hidden_toggle,
            ),
            code="""
            if (ht.active) {
                numinputs[0].value=source_rotmatrix_etc.data['0'][0];
                numinputs[1].value=source_rotmatrix_etc.data['0'][1];
                numinputs[2].value=source_rotmatrix_etc.data['0'][2];
                numinputs[3].value=source_rotmatrix_etc.data['1'][0];
                numinputs[4].value=source_rotmatrix_etc.data['1'][1];
                numinputs[5].value=source_rotmatrix_etc.data['1'][2];
                numinputs[6].value=source_rotmatrix_etc.data['2'][0];
                numinputs[7].value=source_rotmatrix_etc.data['2'][1];
                numinputs[8].value=source_rotmatrix_etc.data['2'][2];
                numinputs[9].value+=1.;
            }
        """,
        )
    )

    def compute_oriented_signature(
        adata, embedding_key, obs_indices_A, dir_x, dir_y, hidden_num_in
    ):
        rotmatrix = np.array([inpt.value for inpt in hidden_num_in[:9]]).reshape(3, 3)
        points_in_A = np.asarray(adata.obsm[embedding_key][obs_indices_A])
        if points_in_A.shape[1] > 2:
            if sl_component1 is None:
                points3d = points_in_A[:, :3]
            else:
                points3d = points_in_A[
                    :,
                    [
                        max(sl_component1.active - 1, 0),
                        max(sl_component2.active - 1, 0),
                        max(sl_component3.active - 1, 0),
                    ],
                ].copy()
                if sl_component1.active == 0:
                    points3d[:, 0] = 0.0
                if sl_component2.active == 0:
                    points3d[:, 1] = 0.0
                if sl_component3.active == 0:
                    points3d[:, 2] = 0.0
        else:
            points3d = np.hstack(
                (points_in_A, np.zeros((len(points_in_A), 1), dtype=np.float32))
            )
        coords = np.dot(rotmatrix, points3d.T).T
        components = dir_x * coords[:, 0] + dir_y * coords[:, 1]
        # corr_scores = []
        # for i in range(adata.n_vars):
        #     a = adata.X[obs_indices_A, i]
        #     try:
        #         res = stats.pearsonr(a, components)
        #     except stats.ConstantInputWarning:
        #         pass
        #     corr_scores.append(np.abs(res.statistic))

        a = pd.DataFrame(adata.X[obs_indices_A, :])
        b = pd.DataFrame(np.tile(components, (a.shape[1], 1)).T)
        corr_scores = a.corrwith(b).to_numpy()
        # corr_scores = np.abs(corr_scores)
        corr_scores = np.vectorize(lambda x: 0.0 if np.isnan(x) else x)(corr_scores)
        sorted_features = np.argsort(np.abs(corr_scores))[::-1][:20]
        cscores = corr_scores[sorted_features]
        corr_dict = dict(map(lambda i, j: (i, j), sorted_features, cscores))
        up_or_down_dict = dict(
            map(lambda i, j: (i, "+" if j >= 0.0 else "-"), sorted_features, cscores)
        )
        return sorted_features, corr_dict, up_or_down_dict

    # TODO remove redundancy (shrink_test is defined twice)
    def shrink_text(s_in, size):
        true_size = max(size, 3)
        if len(s_in) > true_size:
            new_s = ""
            l1 = true_size // 2
            l2 = true_size - l1 - 3
            new_s += s_in[:l1]
            new_s += "..."
            new_s += s_in[-l2:]
        else:
            new_s = s_in
        return new_s

    def oriented_sign_A(
        ad,
        embedding_key,
        arr_layout,
        obs_indices_A,
        dv,
        ms_sign,
        sign_nr,
        hidden_num_in,
    ):
        # print(arr_layout.x_start, arr_layout.y_end, arr_layout.y_start)
        if 0 < len(obs_indices_A) and (
            arr_layout.x_end != arr_layout.x_start
            or arr_layout.y_end != arr_layout.y_start
        ):
            ms_sign.title = "..."
            outputs, corr_dict, up_or_down_dict = compute_oriented_signature(
                ad,
                embedding_key,
                obs_indices_A,
                arr_layout.x_end - arr_layout.x_start,
                arr_layout.y_end - arr_layout.y_start,
                hidden_num_in,
            )
            sign_nr[0] += 1
            dv.text = (
                "Signature #"
                + str(sign_nr[0])
                + ": "
                + ", ".join(['<b>"' + elt + '"</b>' for elt in ad.var_names[outputs]])
            )
            ms_sign.options = [
                (
                    up_or_down_dict[outp] + ad.var_names[outp],
                    up_or_down_dict[outp]
                    + " (Corr.:{:.3f}) ".format(corr_dict[outp])
                    + shrink_text(ad.var_names[outp], 25),
                )
                for outp in outputs
            ]
            ms_sign.title = "Signature #" + str(sign_nr[0])

            unique_labels = []
            unique_labels.append(("[  Subset A  ]", "[  Subset A  ]"))
            unique_labels.append(("[  Rest  ]", "[  Rest  ]"))
            unique_labels += [
                (lbl + ">>yomix>>" + lbl_elt, shrink_text(lbl + " > " + lbl_elt, 35))
                for (lbl, lbl_elt) in ad.uns["all_labels"]
            ]

            # Update label_sign options
            label_sign.options = unique_labels
            label_sign.size = len(label_sign.options)
            # finalize label_sign
            label_sign.title = "Groups"
            label_sign.value = ["[  Subset A  ]", "[  Rest  ]"]

    hidden_numeric_inputs[9].on_change(
        "value",
        lambda attr, old, new: oriented_sign_A(
            adata,
            embedding_key,
            arrow,
            hidden_checkbox_A.active,
            div_signature_list,
            multiselect_signature,
            sign_nr,
            hidden_numeric_inputs,
        ),
    )

    return bt_sign_oriented, help_button_oriented



import numpy as np
import bokeh.models
import bokeh.layouts
import sys
from scipy.stats import rankdata


def signature_buttons(
    adata, offset_text_feature_color, offset_label, hidden_checkbox_A, hidden_checkbox_B
):

    def wasserstein_distance(mu1, sigma1, mu2, sigma2):
        mean_diff = mu1 - mu2
        std_diff = sigma1 - sigma2
        wasserstein = np.sqrt(mean_diff**2 + std_diff**2)
        return wasserstein

    def find_intersection_gaussians(mu1, sigma1, mu2, sigma2):
        # find intersections between two Gaussian probability density functions
        if sigma1 < 0.0 or sigma2 < 0.0:
            sys.exit("Error: negative standard deviation.")

        if sigma2 < 1e-8:
            sigma2 = 1e-8

        if sigma1 < 1e-8:
            sigma1 = 1e-8

        if np.abs(sigma1 - sigma2) < 1e-8:
            x = (mu1 + mu2) / 2
            return x, x

        else:
            sig1square = sigma1**2
            sig2square = sigma2**2
            a = sig2square / 2 - sig1square / 2
            b = sig1square * mu2 - sig2square * mu1
            c = (
                sig2square * mu1**2 / 2
                - sig1square * mu2**2 / 2
                - sig1square * sig2square * (np.log(sigma2) - np.log(sigma1))
            )

            discr = b**2 - 4 * a * c
            if np.abs(discr) < 1e-8:
                discr = 0.0
            x1 = (-b + np.sqrt(discr)) / (2 * a)
            x2 = (-b - np.sqrt(discr)) / (2 * a)

            return x1, x2

    def gaussian_pdf(x, mu, sigma):
        var = sigma**2
        return (
            1.0 / (np.sqrt(2.0 * np.pi * var)) * np.exp(-((x - mu) ** 2) / (2.0 * var))
        )

    def compute_signature(adata, means, stds, obs_indices_A, obs_indices_B=None):
        # STEP 1: sort features using Wasserstein distances

        a2 = adata.X[obs_indices_A, :]
        mu2_array = a2.mean(axis=0)
        sigma2_array = a2.std(axis=0)
        if obs_indices_B is None:
            mu = means
            sigma1_array = stds.to_numpy()
            mu1_array = (
                (mu * adata.n_obs - mu2_array * len(obs_indices_A))
                / (adata.n_obs - len(obs_indices_A))
            ).to_numpy()
        else:
            a1 = adata.X[obs_indices_B, :]
            mu1_array = a1.mean(axis=0)
            sigma1_array = a1.std(axis=0)
        sigma1_array[sigma1_array < 1e-8] = 1e-8
        sigma2_array[sigma2_array < 1e-8] = 1e-8
        dist_list = wasserstein_distance(
            mu1_array, sigma1_array, mu2_array, sigma2_array
        )

        sorted_features = np.argsort(dist_list)[::-1]

        if obs_indices_B is None:
            ref_array = np.arange(adata.n_obs)
            rest_indices = np.arange(adata.n_obs)[~np.in1d(ref_array, obs_indices_A)]
        else:
            rest_indices = obs_indices_B

        samples_A = obs_indices_A
        samples_B = rest_indices

        # Keep only 100 features:
        selected_features = sorted_features[:100]

        def all_mcc(scores1, scores2):
            l1 = scores1.shape[1]
            l2 = scores2.shape[1]

            scores1.sort(axis=1)
            scores2.sort(axis=1)

            all_scores = np.hstack((scores1, scores2))

            ranks = rankdata(all_scores, method="min", axis=1).astype(int)

            all_scores.sort(axis=1)

            ranks1 = ranks[:, :l1]
            ranks2 = ranks[:, l1:]

            def matthews_c(a_, b_, c_, d_, l1_, l2_):

                max_value = np.maximum(l1_, l2_)
                tp = a_ / max_value
                fp = b_ / max_value
                fn = c_ / max_value
                tn = d_ / max_value

                denominator = (tp + fp) * (tp + fn) * (tn + fp) * (tn + fn)

                mcc = (tp * tn - fp * fn) / np.sqrt(denominator)
                return mcc

            def searchsorted2d(a_, b_):
                m, n = a_.shape
                max_num = np.maximum(a_.max(), b_.max()) + 1
                r = max_num * np.arange(a_.shape[0])[:, None]
                p_ = np.searchsorted((a_ + r).ravel(), (b_ + r).ravel()).reshape(m, -1)
                return p_ - n * (np.arange(m)[:, None])

            maxis = np.maximum(np.max(ranks1, axis=1), np.max(ranks2, axis=1))
            rng = (
                np.repeat(np.arange(l1 + l2), scores1.shape[0]).reshape(l1 + l2, -1).T
                + 1
            ).clip(max=maxis[:, None])

            a = np.minimum(searchsorted2d(ranks1, rng)[:, 1:], l1)
            b = l1 - a
            c = np.minimum(searchsorted2d(ranks2, rng)[:, 1:], l2)
            d = l2 - c

            results = matthews_c(a, b, c, d, l1, l2)

            idx = l1 + l2 - 2 - np.abs(results[:, ::-1]).argmax(axis=1)

            first_axis_range = (np.arange(scores1.shape[0]),)
            mccscores = results[first_axis_range, idx]
            # cuts = (
            #     all_scores[first_axis_range, idx] +
            #     all_scores[first_axis_range, idx + 1]
            #     ) / 2.
            # return mccscores, cuts
            return mccscores

        sc1 = adata[samples_A, selected_features].copy().X.T
        sc2 = adata[samples_B, selected_features].copy().X.T
        mccs = all_mcc(sc1, sc2)
        new_selected_features = selected_features[
            np.argsort(np.abs(mccs.flatten()))[::-1]
        ]
        mcc_dict = dict(map(lambda i, j: (i, j), selected_features, mccs.flatten()))
        mcc_dict_abs = dict(
            map(lambda i, j: (i, j), selected_features, np.abs(mccs).flatten())
        )
        new_selected_features = new_selected_features[:20]
        up_or_down_d = {
            ft: ("-" if mcc_dict[ft] > 0.0 else "+") for ft in new_selected_features
        }
        return new_selected_features, mcc_dict_abs, up_or_down_d

    def shrink_text(s_in, size):
        true_size = max(size, 3)
        if len(s_in) > true_size:
            new_s = ""
            l1 = true_size // 2
            l2 = true_size - l1 - 3
            new_s += s_in[:l1]
            new_s += "..."
            new_s += s_in[-l2:]
        else:
            new_s = s_in
        return new_s

    def sign_A_vs_rest(ad, obs_indices, dv, ms_sign, sign_nr, label_sign):
        if len(obs_indices) > 0 and len(obs_indices) < ad.n_obs:
            ms_sign.title = "..."
            label_sign.title = "..."
            outputs, mcc_dict, up_or_down_dict = compute_signature(
                ad,
                ad.var["mean_values_local_yomix"],
                ad.var["standard_deviations_local_yomix"],
                obs_indices,
                None,
            )
            sign_nr[0] += 1
            dv.text = (
                "Signature #"
                + str(sign_nr[0])
                + ": "
                + ", ".join(['<b>"' + elt + '"</b>' for elt in ad.var_names[outputs]])
            )
            ms_sign.options = [
                (
                    up_or_down_dict[outp] + ad.var_names[outp],
                    up_or_down_dict[outp]
                    + " (MCC:{:.3f}) ".format(mcc_dict[outp])
                    + shrink_text(ad.var_names[outp], 25),
                )
                for outp in outputs
            ]
            ms_sign.title = "Signature #" + str(sign_nr[0])

            unique_labels = []
            unique_labels.append(("[  Subset A  ]", "[  Subset A  ]"))
            unique_labels.append(("[  Rest  ]", "[  Rest  ]"))
            unique_labels += [
                (lbl + ">>yomix>>" + lbl_elt, shrink_text(lbl + " > " + lbl_elt, 35))
                for (lbl, lbl_elt) in ad.uns["all_labels"]
            ]

            # Update label_sign options
            label_sign.options = unique_labels
            label_sign.size = len(label_sign.options)
            # finalize label_sign
            label_sign.title = "Groups"
            label_sign.value = ["[  Subset A  ]", "[  Rest  ]"]

    def sign_A_vs_B(ad, obs_indices_A, obs_indices_B, dv, ms_sign, sign_nr, label_sign):
        if (
            len(obs_indices_A) > 0
            and len(obs_indices_A) < ad.n_obs
            and len(obs_indices_B) > 0
            and len(obs_indices_B) < ad.n_obs
        ):
            ms_sign.title = "..."
            label_sign.title = "..."
            outputs, mcc_dict, up_or_down_dict = compute_signature(
                ad,
                ad.var["mean_values_local_yomix"],
                ad.var["standard_deviations_local_yomix"],
                obs_indices_A,
                obs_indices_B,
            )
            sign_nr[0] += 1
            dv.text = (
                "Signature #"
                + str(sign_nr[0])
                + ": "
                + ", ".join(['<b>"' + elt + '"</b>' for elt in ad.var_names[outputs]])
            )
            ms_sign.options = [
                (
                    up_or_down_dict[outp] + ad.var_names[outp],
                    up_or_down_dict[outp]
                    + " (MCC:{:.3f}) ".format(mcc_dict[outp])
                    + shrink_text(ad.var_names[outp], 25),
                )
                for outp in outputs
            ]
            ms_sign.title = "Signature #" + str(sign_nr[0])

            unique_labels = []
            unique_labels.append(("[  Subset A  ]", "[  Subset A  ]"))
            unique_labels.append(("[  Subset B  ]", "[  Subset B  ]"))
            unique_labels += [
                (lbl + ">>yomix>>" + lbl_elt, shrink_text(lbl + " > " + lbl_elt, 35))
                for (lbl, lbl_elt) in ad.uns["all_labels"]
            ]

            # Update label_sign options
            label_sign.options = unique_labels
            label_sign.size = len(label_sign.options)

            # finalize label_sign
            label_sign.title = "Groups"
            label_sign.value = ["[  Subset A  ]", "[  Subset B  ]"]

    div_signature_list = bokeh.models.Div(
        width=235, height=50, height_policy="fixed", text="Signature #0:"
    )
    signature_nr = [0]
    options = []
    multiselect_signature = bokeh.models.MultiSelect(
        title="Signature #0",
        options=options,
        width=235,
        max_width=235,
        size=20,
        width_policy="max",
    )

    options = []
    label_signature = bokeh.models.MultiSelect(
        title="Groups",
        options=options,
        width=235,
        max_width=235,
        width_policy="max",
    )

    def multiselect_function(feature_list):
        of_text = ""
        for i in range(len(feature_list)):
            if feature_list[i][0] == "+":
                if i == 0:
                    of_text += feature_list[i][1:]
                else:
                    of_text += "  +  " + feature_list[i][1:]
            else:  # feature_list[i][0] == "-"
                of_text += "  -  " + feature_list[i][1:]
        offset_text_feature_color.value = of_text

    multiselect_signature.on_change(
        "value", lambda attr, old, new: multiselect_function(new)
    )

    def label_function(feature_list):
        of_text = ""
        for i in range(len(feature_list)):
            of_text += feature_list[i] + "//yomix//"
        offset_label.value = of_text

    label_signature.on_change("value", lambda attr, old, new: label_function(new))

    tooltip1 = bokeh.models.Tooltip(
        content="Requires setting subset A.\u00A0\u00A0", position="right"
    )
    help_button1 = bokeh.models.HelpButton(tooltip=tooltip1, margin=(3, 0, 3, 0))
    bt_sign1 = bokeh.models.Button(
        label="Compute signature (A vs. rest)", width=190, margin=(5, 0, 5, 5)
    )

    bt_sign1.on_click(
        lambda event: sign_A_vs_rest(
            adata,
            hidden_checkbox_A.active,
            div_signature_list,
            multiselect_signature,
            signature_nr,
            label_signature,
        )
    )

    tooltip2 = bokeh.models.Tooltip(
        content="Requires setting subsets A and B.\u00A0\u00A0", position="right"
    )
    help_button2 = bokeh.models.HelpButton(tooltip=tooltip2, margin=(3, 0, 3, 0))
    bt_sign2 = bokeh.models.Button(
        label="Compute signature (A vs. B)", width=190, margin=(5, 0, 5, 5)
    )

    bt_sign2.on_click(
        lambda event: sign_A_vs_B(
            adata,
            hidden_checkbox_A.active,
            hidden_checkbox_B.active,
            div_signature_list,
            multiselect_signature,
            signature_nr,
            label_signature,
        )
    )

    return (
        bt_sign1,
        bt_sign2,
        help_button1,
        help_button2,
        multiselect_signature,
        div_signature_list,
        signature_nr,
        label_signature,
    )



from . import plotting, tools
import sys
from pathlib import Path

__version__ = Path(__file__).with_name("VERSION").read_text().strip()

sys.modules.update({f"{__name__}.{m}": globals()[m] for m in ["plotting", "tools"]})



"""# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'yomix'
copyright = '2024, nicolas perrin gilbert'
author = 'nicolas perrin gilbert'
release = '0.0.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']
"""

# Licensed under the BSD 3-Clause License.

import sys
from os.path import abspath

sys.path.insert(0, abspath(".."))

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx-mathjax-offline",
    "sphinx.ext.napoleon",  # before sphinx_autodoc_typehints
    "sphinx_autodoc_typehints",
    "sphinx.ext.autosummary",  # summary table
]

# List of modules to be mocked up
autodoc_mock_imports = []

# Configure sphinx.ext.autosummary
autosummary_generate = True
autoclass_content = "both"  # add __init__ docstrings to class summaries
html_show_sourcelink = False  # remove "view source code" from top of page
autodoc_inherit_docstrings = True  # if no docstring, inherit from base class
set_type_checking_flag = True  # 'expensive' imports (sphinx_autodoc_typehints)
nbsphinx_allow_errors = True  # continue through Jupyter errors
add_module_names = False  # remove namespaces from class/method signatures

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "yomix"
copyright = "2024 "
author = "Nicolas Perrin-Gilbert"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ""

# The full version, including alpha/beta/rc tags.
release = ""

""" #Read version info directly from the module's __init__.py
init_path = join(dirname(dirname(str(abspath(__file__)))), "yomix")
with open(f"{init_path}/__init__.py", "r") as fh:
    for line in fh:
        match = re.match('__version__ = "((\\d.\\d).\\d)[a-z0-9\\-]*".*', line)
        if match is not None:
            release = match.group(1)
            version = match.group(2)
            break"""

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_templates"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {}

# Output file base name for HTML help builder.
htmlhelp_basename = "yomixdoc"



import bokeh.models


def subset_buttons(points_bokeh_plot, source_rotmatrix_etc):
    source = points_bokeh_plot.select(dict(name="scatterplot"))[0].data_source
    button_width = 112
    # big_button_width = 235

    toggle_A = bokeh.models.Toggle(
        label="Highlight A: ON", width=button_width, active=True
    )
    toggle_B = bokeh.models.Toggle(
        label="Highlight B: ON", width=button_width, active=True
    )

    hidden_checkbox_A = bokeh.models.CheckboxGroup(
        labels=["0"] * len(source.data["name"]), active=[]
    )

    hidden_checkbox_B = bokeh.models.CheckboxGroup(
        labels=["0"] * len(source.data["name"]), active=[]
    )

    size_redef_code = """
        point_s[i] = point_s_ref[i] * size_coef[0] * (
            2.5 * Math.max(hA[i], hB[i]) + 1.0);
        line_c[i] = -1. + hA[i] * 0.25 + hB[i] * 0.90;
        line_w[i] = point_s_ref[i] * size_coef[0] * 0.85 * Math.max(
            hA[i], hB[i]);
    """

    toggle_A.js_on_change(
        "active",
        bokeh.models.CustomJS(
            args=dict(source=source, source_rotmatrix_etc=source_rotmatrix_etc),
            code="""
        const data = source.data;
        const data_rotmatrix_etc = source_rotmatrix_etc.data;
        const hA = data['highlighted_A'];
        const hB = data['highlighted_B'];
        const sA = data['subset_A'];
        const point_s = data['point_size'];
        const point_s_ref = data['point_size_ref'];
        const line_c = data['line_color'];
        const line_w = data['line_width'];
        const size_coef = data_rotmatrix_etc['size_coef'];
        if (this.active) {
            this.label = 'Highlight A: ON';
            for (let i = 0; i < hA.length; i++) {
                hA[i] = sA[i];"""
            + size_redef_code
            + """
            }
        } else {
            this.label = 'Highlight A: OFF';
            for (let i = 0; i < hA.length; i++) {
                hA[i] = 0.;"""
            + size_redef_code
            + """
            }
        }
        source.change.emit();
    """,
        ),
    )
    toggle_B.js_on_change(
        "active",
        bokeh.models.CustomJS(
            args=dict(source=source, source_rotmatrix_etc=source_rotmatrix_etc),
            code="""
        const data = source.data;
        const data_rotmatrix_etc = source_rotmatrix_etc.data;
        const hB = data['highlighted_B'];
        const hA = data['highlighted_A'];
        const sB = data['subset_B'];
        const point_s = data['point_size'];
        const point_s_ref = data['point_size_ref'];
        const line_c = data['line_color'];
        const line_w = data['line_width'];
        const size_coef = data_rotmatrix_etc['size_coef'];
        if (this.active) {
            this.label = 'Highlight B: ON';
            for (let i = 0; i < hB.length; i++) {
                hB[i] = sB[i];"""
            + size_redef_code
            + """
            }
        } else {
            this.label = 'Highlight B: OFF';
            for (let i = 0; i < hB.length; i++) {
                hB[i] = 0.;"""
            + size_redef_code
            + """
            }
        }
        source.change.emit();
    """,
        ),
    )

    callback_AB_part_1 = """
        const indices = source.selected.indices;
        const data = source.data;
        const data_rotmatrix_etc = source_rotmatrix_etc.data;
        const point_s = data['point_size'];
        const point_s_ref = data['point_size_ref'];
        const sA = data['subset_A'];
        const sB = data['subset_B'];
        const hA = data['highlighted_A'];
        const hB = data['highlighted_B'];
        const line_c = data['line_color'];
        const line_w = data['line_width'];
        const size_coef = data_rotmatrix_etc['size_coef'];
        for (let i = 0; i < point_s.length; i++) {
    """
    callback_AB_part_2 = (
        size_redef_code
        + """
        }

        source.change.emit();
        toggle.active=true;
    """
    )

    callback_js_A = bokeh.models.CustomJS(
        args=dict(
            source=source,
            source_rotmatrix_etc=source_rotmatrix_etc,
            toggle=toggle_A,
            cbx=hidden_checkbox_A,
            cbx_other=hidden_checkbox_B,
        ),
        code=callback_AB_part_1
        + """
            sA[i] = 0.;
        }
        var t = [];
        var t_remove = [];
        for (let i = 0; i < indices.length; i++) {
            sA[indices[i]] = 1.;
            if (sB[indices[i]] == 1.) {
                t_remove.push(data["index"][indices[i]]);
            }
            sB[indices[i]] = 0.;
            hB[indices[i]] = 0.;
            t.push(data["index"][indices[i]]);
        }
        if (t_remove.length > 0) {
            let difference = cbx_other.active.filter(x => !t_remove.includes(x));
            cbx_other.active = difference;
            cbx_other.change.emit();
        }
        cbx.active = t;
        cbx.change.emit();
        const t_active = toggle.active;
        for (let i = 0; i < point_s.length; i++) {
            if (t_active) {
                hA[i] = sA[i];
            }
        """
        + callback_AB_part_2,
    )

    callback_js_B = bokeh.models.CustomJS(
        args=dict(
            source=source,
            source_rotmatrix_etc=source_rotmatrix_etc,
            toggle=toggle_B,
            cbx=hidden_checkbox_B,
            cbx_other=hidden_checkbox_A,
        ),
        code=callback_AB_part_1
        + """
            sB[i] = 0.;
        }
        var t = [];
        var t_remove = [];
        for (let i = 0; i < indices.length; i++) {
            sB[indices[i]] = 1.;
            if (sA[indices[i]] == 1.) {
                t_remove.push(data["index"][indices[i]]);
            }
            sA[indices[i]] = 0.;
            hA[indices[i]] = 0.;
            t.push(data["index"][indices[i]]);
        }
        if (t_remove.length > 0) {
            let difference = cbx_other.active.filter(x => !t_remove.includes(x));
            cbx_other.active = difference;
            cbx_other.change.emit();
        }
        cbx.active = t;
        cbx.change.emit();
        const t_active = toggle.active;
        for (let i = 0; i < point_s.length; i++) {
            if (t_active) {
                hB[i] = sB[i];
            }
        """
        + callback_AB_part_2,
    )

    bt_A = bokeh.models.Button(
        label="Set subset A", width=button_width, button_type="primary"
    )
    bt_B = bokeh.models.Button(
        label="Set subset B", width=button_width, button_type="danger"
    )

    bt_A.js_on_click(callback_js_A)
    bt_B.js_on_click(callback_js_B)

    bt_AplusB = bokeh.models.Button(label="Select A+B", width=button_width)
    bt_AplusB.js_on_click(
        bokeh.models.CustomJS(
            args=dict(source=source),
            code="""
        const data = source.data;
        var t = [];
        const sA = data['subset_A'];
        const sB = data['subset_B'];
        const point_s = data['point_size'];
        for (let i = 0; i < point_s.length; i++) {
            if (sB[i] == 1. || sA[i] == 1.) {
                t.push(i);
            }
        }
        source.selected.indices = t;
        source.change.emit();
    """,
        )
    )

    bt_nothing = bokeh.models.Button(label="Select nothing", width=button_width)
    bt_nothing.js_on_click(
        bokeh.models.CustomJS(
            args=dict(source=source),
            code="""
        const data = source.data;
        source.selected.indices = [];
        source.change.emit();
    """,
        )
    )

    return (
        bt_A,
        toggle_A,
        hidden_checkbox_A,
        bt_B,
        toggle_B,
        hidden_checkbox_B,
        bt_AplusB,
        bt_nothing,
    )



import numpy as np
import bokeh.models
import bokeh.io
import bokeh.layouts
import re
from bokeh.models import (
    HoverTool,
    LassoSelectTool,
    WheelZoomTool,
    ResetTool,
    HelpTool,
    InlineStyleSheet,
    Range1d,
)
from bokeh.models.tools import CustomAction
import bokeh.plotting as bkp
import matplotlib
import matplotlib.pyplot as plt
from pathlib import Path

MAX_UNIQUE_VALUES = 70
MAX_UNIQUE_RATIO = 0.5


def check_obs_field(xd, field):
    unique_items_nr = len(xd.obs[field].unique())
    return (
        unique_items_nr < MAX_UNIQUE_VALUES
        and unique_items_nr / xd.n_obs <= MAX_UNIQUE_RATIO
    )


def main_figure(adata, embedding_key, width=900, height=600, title=""):

    obs_string = [
        x
        for x in list(adata.obs.select_dtypes(include=["category", "object"]).keys())
        if check_obs_field(adata, x)
    ]
    obs_numerical = list(adata.obs.select_dtypes(np.number).keys())

    embedding_size = adata.obsm[embedding_key].shape[1]

    if embedding_size > 3:
        higher_dim = True
    else:
        higher_dim = False

    tooltip_list = []
    data = {
        "x": np.asarray(adata.obsm[embedding_key][:, 0], dtype=np.float32).copy(),
        "y": np.asarray(adata.obsm[embedding_key][:, 1], dtype=np.float32).copy(),
        "color": np.zeros(len(adata.obsm[embedding_key]), dtype=np.float32),
        "index": np.zeros(len(adata.obsm[embedding_key]), dtype=int),
        "name": np.full(len(adata.obsm[embedding_key]), "", dtype=object),
    }
    if embedding_size >= 3:
        data["z"] = np.asarray(adata.obsm[embedding_key][:, 2], dtype=np.float32).copy()
    else:
        data["z"] = np.zeros_like(data["x"], dtype=np.float32).copy()
    for i in range(len(data["x"])):
        data["index"][i] = i
        data["name"][i] = adata.obs_names[i]
    # tooltip_list.append("index")
    tooltip_list.append("name")
    for key in adata.obs.keys():
        if key not in ("name", "index"):
            data[key] = np.asarray(adata.obs[key])
            tooltip_list.append(key)
    for i in range(embedding_size):
        data["original_embedding_" + str(i)] = np.asarray(
            adata.obsm[embedding_key][:, i], dtype=np.float32
        )

    if title == "":
        title = (
            "Yomix " + Path(__file__).parent.with_name("VERSION").read_text().strip()
        )
    xyz_vectors = np.vstack((data["x"], data["y"], data["z"]))
    true_center = np.array([[data["x"].mean(), data["y"].mean(), data["z"].mean()]]).T
    xyz_norms = np.linalg.norm(xyz_vectors - true_center, axis=0)
    xyz_max = xyz_norms.max()
    x_min = data["x"].min()
    x_max = data["x"].max()
    y_min = data["y"].min()
    y_max = data["y"].max()

    data["point_size"] = data["z"].copy()
    for i in range(len(data["point_size"])):
        data["point_size"][i] = 5.0 / (
            1.0
            + np.exp(
                min(
                    max(
                        (-data["point_size"][i] + true_center[2, 0])
                        / (xyz_max + 0.0000001),
                        -1.0,
                    ),
                    1.0,
                )
            )
        )

    data["x_saved"] = data["x"].copy()
    data["y_saved"] = data["y"].copy()
    data["z_saved"] = data["z"].copy()
    data["point_size_ref"] = data["point_size"].copy()
    data["color_ref"] = data["color"].copy()

    data["subset_A"] = np.zeros(len(data["point_size"]), dtype=np.float32)
    data["subset_B"] = np.zeros(len(data["point_size"]), dtype=np.float32)
    data["highlighted_A"] = np.zeros(len(data["point_size"]), dtype=np.float32)
    data["highlighted_B"] = np.zeros(len(data["point_size"]), dtype=np.float32)
    data["line_color"] = -np.ones(len(data["point_size"]), dtype=np.float32)
    data["line_width"] = np.zeros(len(data["point_size"]), dtype=np.float32)

    source = bokeh.models.ColumnDataSource(data=data)

    points_bokeh_plot = bkp.figure(
        width=width,
        height=height,
        title=title,
        x_axis_type="linear",
        y_axis_type="linear",
        toolbar_location="left",
    )
    points_bokeh_plot.xgrid.grid_line_color = None
    points_bokeh_plot.ygrid.grid_line_color = None
    points_bokeh_plot.xaxis.axis_label = ""
    points_bokeh_plot.yaxis.axis_label = ""
    points_bokeh_plot.xaxis.major_tick_line_color = None  # turn off x-axis major ticks
    points_bokeh_plot.xaxis.minor_tick_line_color = None  # turn off x-axis minor ticks
    points_bokeh_plot.yaxis.major_tick_line_color = None  # turn off y-axis major ticks
    points_bokeh_plot.yaxis.minor_tick_line_color = None  # turn off y-axis minor ticks
    points_bokeh_plot.xaxis.major_label_text_font_size = (
        "0pt"  # preferred method for removing tick labels
    )
    points_bokeh_plot.yaxis.major_label_text_font_size = (
        "0pt"  # preferred method for removing tick labels
    )
    points_bokeh_plot.x_range = Range1d(x_min, x_max)
    points_bokeh_plot.y_range = Range1d(y_min, y_max)

    nipy_spectral_colormap = plt.get_cmap("nipy_spectral")
    nipy_spectral_colormap._segmentdata["red"][-1] = (1.0, 0.9, 0.9)
    nipy_spectral_colormap._segmentdata["green"][-1] = (1.0, 0.5, 0.5)
    nipy_spectral_colormap._segmentdata["blue"][-1] = (1.0, 0.5, 0.5)
    nipy_spectral_colormap._init()
    nipy_colors = [
        matplotlib.colors.rgb2hex(plt.get_cmap("nipy_spectral")(i)) for i in range(256)
    ]
    viridis_colors = list(bokeh.palettes.Viridis256)
    custom_color_mapper = bokeh.models.LinearColorMapper(
        palette=nipy_colors + viridis_colors, low=-1.0, high=1.0
    )

    color_arg = {"field": "color", "transform": custom_color_mapper}
    line_color_arg = {"field": "line_color", "transform": custom_color_mapper}

    scatter_scatterplot = points_bokeh_plot.scatter(
        "x",
        "y",
        source=source,
        marker="circle",
        fill_alpha=1.0,
        size="point_size",
        color=color_arg,
        line_color=line_color_arg,
        line_width="line_width",
        line_alpha=1.0,
        name="scatterplot",
    )

    if embedding_size >= 3:
        tooltips = [("coordinates", "[X: @{x_saved}, Y: @{y_saved}, Z: @{z_saved}]")]
    else:
        tooltips = [("coordinates", "[X: @{x_saved}, Y: @{y_saved}]")]
    for elt in tooltip_list:
        tooltips.append((elt, "@{" + elt + "}"))
    hover = HoverTool(tooltips=tooltips)
    points_bokeh_plot.add_tools(hover)
    lasso = LassoSelectTool()
    points_bokeh_plot.add_tools(lasso)
    lasso_select_tool = points_bokeh_plot.select(dict(type=LassoSelectTool))
    lasso_select_tool.continuous = False
    points_bokeh_plot.toolbar.active_drag = points_bokeh_plot.select_one(
        LassoSelectTool
    )
    points_bokeh_plot.toolbar.active_scroll = points_bokeh_plot.select_one(
        WheelZoomTool
    )

    reset = points_bokeh_plot.select(dict(type=ResetTool))
    reset.visible = False
    helpt = points_bokeh_plot.select(dict(type=HelpTool))
    helpt.visible = False

    scatter_scatterplot.nonselection_glyph.fill_alpha = 0.08
    scatter_scatterplot.nonselection_glyph.line_alpha = 0.0
    scatter_scatterplot.nonselection_glyph.fill_color = "gray"

    hover.renderers = [scatter_scatterplot]
    lasso.renderers = [scatter_scatterplot]

    rotmatrix_etc = {
        "0": [1.0, 0.0, 0.0],
        "1": [0.0, 1.0, 0.0],
        "2": [0.0, 0.0, 1.0],
        "delta": [0.0, 0.0, 0.0],
        "center": true_center.flatten().tolist(),
        "offset_angle": [0.0, 0.0, 0.0],
        "max_norm": [xyz_max, -1.0, -1.0],
        "xrange": [x_min, x_max, -1.0],
        "yrange": [y_min, y_max, -1.0],
        "size_coef": [1.0, -1.0, -1.0],
        "width": [
            max(
                points_bokeh_plot.width
                - 41.0,  # 41 is an estimate of the toolbar width
                1.0,
            ),
            -1.0,
            -1.0,
        ],
        "legend_width": [0.0, -1.0, -1.0],
        "height": [
            max(
                points_bokeh_plot.height
                - 31.0,  # 31 is an estimate of the title height
                1.0,
            ),
            -1.0,
            -1.0,
        ],
    }

    source_rotmatrix_etc = bokeh.models.ColumnDataSource(data=rotmatrix_etc)

    tool = CustomAction(
        description="Reset view to selected points",
        icon="reset",
        callback=bokeh.models.CustomJS(
            args=dict(pbp=points_bokeh_plot, source=source),
            code="""
        const indices = source.selected.indices;
        if (indices.length > 0) {
            const subarray_x = indices.map(i => source.data["x"][i]);
            const subarray_y = indices.map(i => source.data["y"][i]);
            pbp.x_range.start = Math.min.apply(null, subarray_x) - 0.00000001;
            pbp.x_range.end = Math.max.apply(null, subarray_x) + 0.00000001;
            pbp.y_range.start = Math.min.apply(null, subarray_y) - 0.00000001;
            pbp.y_range.end = Math.max.apply(null, subarray_y) + 0.00000001;
        } else {
            const subarray_x = source.data["x"];
            const subarray_y = source.data["y"];
            pbp.x_range.start = Math.min.apply(null, subarray_x) - 0.00000001;
            pbp.x_range.end = Math.max.apply(null, subarray_x) + 0.00000001;
            pbp.y_range.start = Math.min.apply(null, subarray_y) - 0.00000001;
            pbp.y_range.end = Math.max.apply(null, subarray_y) + 0.00000001;
        }
    """,
        ),
    )
    points_bokeh_plot.add_tools(tool)

    callback_js_range_change = []

    for elt in [
        "['xrange'][0] = cb_obj.start;",
        "['xrange'][1] = cb_obj.end;",
        "['yrange'][0] = cb_obj.start;",
        "['yrange'][1] = cb_obj.end;",
    ]:
        callback_js_range_change.append(
            bokeh.models.CustomJS(
                args=dict(source_rotmatrix_etc=source_rotmatrix_etc),
                code="""
                const data_rotmatrix_etc = source_rotmatrix_etc.data;
                data_rotmatrix_etc
            """
                + elt,
            )
        )

    points_bokeh_plot.x_range.js_on_change("start", callback_js_range_change[0])
    points_bokeh_plot.x_range.js_on_change("end", callback_js_range_change[1])
    points_bokeh_plot.y_range.js_on_change("start", callback_js_range_change[2])
    points_bokeh_plot.y_range.js_on_change("end", callback_js_range_change[3])

    div_sample_names = bokeh.models.Div(
        width=400, height=50, height_policy="fixed", text="Sample name(s): "
    )
    cb = bokeh.models.CustomJS(
        args=dict(hvr=hover, div=div_sample_names, source=source, col_name="name"),
        code="""
            if (cb_data['index'].indices.length > 0) {
                const data = source.data;
                const line_list = [];
                for (let i = 0; i<cb_data['index'].indices.length; i++) {
                    var line = '<b>"'
                    line += data[col_name][cb_data['index'].indices[i]] + '"</b>'
                    line_list.push(line)
                }
                div.text = "Sample name(s): " + line_list.join(", ");
            }
        """,
    )
    hover.callback = cb  # callback whenever the HoverTool function is called

    callback_js_selected_indices_change = bokeh.models.CustomJS(
        args=dict(
            source=source,
            source_rotmatrix_etc=source_rotmatrix_etc,
            div=div_sample_names,
        ),
        code="""
            const data_rotmatrix_etc = source_rotmatrix_etc.data;
            const rc = data_rotmatrix_etc['center'];
            const data = source.data;
            const x = data['x'];
            const y = data['y'];
            const z = data['z'];
            const indices = source.selected.indices;
            var sum_x = 0.;
            var sum_y = 0.;
            var sum_z = 0.;
            if (indices.length > 0) {
                for (let i = 0; i < indices.length; i++) {
                    sum_x = sum_x + x[indices[i]];
                    sum_y = sum_y + y[indices[i]];
                    sum_z = sum_z + z[indices[i]];
                }
                rc[0] = sum_x / indices.length;
                rc[1] = sum_y / indices.length;
                rc[2] = sum_z / indices.length;
            } else {
                for (let i = 0; i < x.length; i++) {
                    sum_x = sum_x + x[i];
                    sum_y = sum_y + y[i];
                    sum_z = sum_z + z[i];
                }
                rc[0] = sum_x / x.length;
                rc[1] = sum_y / x.length;
                rc[2] = sum_z / x.length;
            }
    """,
    )

    source.selected.js_on_change("indices", callback_js_selected_indices_change)

    stylesheet = InlineStyleSheet(
        css=".noUi-vertical .noUi-origin {top: 0%;} .noUi-base .noUi-connects "
        "{height: 100px;} .noUi-target.noUi-vertical {margin: 10px 0px 0px 25px;}"
    )

    stylesheet2 = InlineStyleSheet(
        css=".noUi-vertical .noUi-origin {top: 0%;} .noUi-base .noUi-connects "
        "{height: "
        + str(height - 55)
        + "px;} .noUi-target.noUi-vertical {margin: 32px 0px 0px -55px;}"
    )

    bt_slider_point_size = bokeh.models.Slider(
        start=0.0,
        end=10.0,
        value=3.0,
        step=0.05,
        title="Point size",
        orientation="vertical",
        stylesheets=[stylesheet],
        width=100,
        show_value=False,
        direction="rtl",
    )

    bt_hidden_slider_yaw = bokeh.models.Slider(
        start=-360.0,
        end=360.0,
        value=0.0,
        step=4.0,
        title="Rotate (yaw)",
        width=100,
        show_value=False,
        name="bt_hidden_slider_yaw",
        visible=False,
    )

    bt_slider_range = bokeh.models.RangeSlider(
        start=0.0,
        end=1.0,
        value=(0.0, 1.0),
        step=0.01,
        orientation="vertical",
        stylesheets=[stylesheet2],
        width=100,
        show_value=False,
        direction="rtl",
        visible=False,
    )

    callback_js_slider_range = bokeh.models.CustomJS(
        args=dict(source=source),
        code="""
        const maxv = cb_obj.end;
        const minv = cb_obj.start;
        const spread = maxv - minv;
        const value_low = (cb_obj.value[0] - minv) / spread;
        const value_high = (cb_obj.value[1] - minv) / spread;
        const data = source.data;
        var t = [];
        const color = data['color'];
        for (let i = 0; i < color.length; i++) {
            if (color[i] >= value_low && color[i] <= value_high) {
                t.push(i);
            }
        }
        source.selected.indices = t;
        source.change.emit();
    """,
    )

    bt_slider_range.js_on_change("value_throttled", callback_js_slider_range)

    bt_toggle_anim = bokeh.models.Toggle(
        label="Animation: OFF", width=100, active=False, name="bt_toggle_anim"
    )

    bt_toggle_anim.js_on_change(
        "active",
        bokeh.models.CustomJS(
            code="""
            if (this.active) {
                this.label = 'Animation: ON';
            } else {
                this.label = 'Animation: OFF';
            }
        """,
        ),
    )

    bt_slider_yaw = bokeh.models.Slider(
        start=-360.0,
        end=360.0,
        value=0.0,
        step=4.0,
        title="Rotate (yaw)",
        width=100,
        show_value=False,
    )

    bt_slider_pitch = bokeh.models.Slider(
        start=-360.0,
        end=360.0,
        value=0.0,
        step=4.0,
        title="Rotate (pitch)",
        orientation="vertical",
        stylesheets=[stylesheet],
        width=100,
        show_value=False,
    )

    bt_slider_roll = bokeh.models.Slider(
        start=-360.0,
        end=360.0,
        value=0.0,
        step=4.0,
        title="Rotate (roll)",
        width=100,
        show_value=False,
    )

    if higher_dim:
        sl_component1 = bokeh.models.RadioButtonGroup(
            labels=["None"] + ["d" + str(i + 1) for i in range(embedding_size)],
            active=1,
            margin=[0, 0, 0, 0],
        )
        sl_component2 = bokeh.models.RadioButtonGroup(
            labels=["None"] + ["d" + str(i + 1) for i in range(embedding_size)],
            active=2,
            margin=[0, 0, 0, 0],
        )
        sl_component3 = bokeh.models.RadioButtonGroup(
            labels=["None"] + ["d" + str(i + 1) for i in range(embedding_size)],
            active=3,
            margin=[0, 0, 0, 0],
        )

    size_redef_code = """
        point_s[i] = point_s_ref[i] * size_coef[0] * (
            2.5 * Math.max(hA[i], hB[i]) + 1.0);
        line_c[i] = -1. + hA[i] * 0.25 + hB[i] * 0.90;
        line_w[i] = point_s_ref[i] * size_coef[0] * 0.85 * Math.max(
            hA[i], hB[i]);
    """

    callback_js_point_size = bokeh.models.CustomJS(
        args=dict(source=source, source_rotmatrix_etc=source_rotmatrix_etc),
        code="""
        const f = cb_obj.value;
        const data_rotmatrix_etc = source_rotmatrix_etc.data;
        const size_coef = data_rotmatrix_etc['size_coef'];
        const data = source.data;
        const point_s = data['point_size'];
        const point_s_ref = data['point_size_ref'];
        const hA = data['highlighted_A'];
        const hB = data['highlighted_B'];
        const line_c = data['line_color'];
        const line_w = data['line_width'];
        size_coef[0] = (f/3.)**2;
        for (let i = 0; i < point_s.length; i++) {"""
        + size_redef_code
        + """
        }
        source_rotmatrix_etc.change.emit();
        source.change.emit();
    """,
    )

    bt_slider_point_size.js_on_change("value", callback_js_point_size)

    code_part1_0 = """
        function matrixMultiply(matrix1, matrix2) {
            var result = [];
            var rows1 = matrix1.length;
            var cols1 = matrix1[0].length;
            var rows2 = matrix2.length;
            var cols2 = matrix2[0].length;

            if (cols1 !== rows2) {
                // Matrix multiplication is not possible
                throw new Error("Incompatible matrix dimensions.");
            }

            for (var i = 0; i < rows1; i++) {
                result[i] = [];
                for (var j = 0; j < cols2; j++) {
                    var sum = 0;
                    for (var k = 0; k < cols1; k++) {
                        sum += matrix1[i][k] * matrix2[k][j];
                    }
                    result[i][j] = sum;
                }
            }
            return result;
        }

        function normalize(vec) {
            // Function to normalize a vector
            var norm = Math.sqrt(vec.reduce((acc, val) => acc + val ** 2, 0));
            return vec.map(val => val / norm);
        }

        function rotationMatrix(axis, angle) {
            /**
            * Compute the rotation matrix around a given axis by a given angle.
            *
            * Parameters:
            *     axis: A 3-element array representing the rotation axis.
            *     angle: The rotation angle in radians.
            *
            * Returns:
            *     A 3x3 rotation matrix.
            */
            axis = normalize(axis); // Normalize the axis vector

            var ux = axis[0], uy = axis[1], uz = axis[2];
            var cosTheta = Math.cos(angle);
            var sinTheta = Math.sin(angle);
            var rotMatrix = [
                [
                    cosTheta + ux ** 2 * (1 - cosTheta),
                    ux * uy * (1 - cosTheta) - uz * sinTheta,
                    ux * uz * (1 - cosTheta) + uy * sinTheta
                ],
                [
                    uy * ux * (1 - cosTheta) + uz * sinTheta,
                    cosTheta + uy ** 2 * (1 - cosTheta),
                    uy * uz * (1 - cosTheta) - ux * sinTheta
                ],
                [
                    uz * ux * (1 - cosTheta) - uy * sinTheta,
                    uz * uy * (1 - cosTheta) + ux * sinTheta,
                    cosTheta + uz ** 2 * (1 - cosTheta)
                ]
            ];
            return rotMatrix;
        }

        function scaleTransform(matrix, scale_0, scale_1, scale_2) {
            var matrix1 = [
                [
                    scale_0 /Math.sqrt(scale_1), 0., 0.
                ],
                [
                    0., Math.sqrt(scale_1), 0.
                ],
                [
                    0., 0., scale_2
                ]
            ];
            var matrix2 = [
                [
                     Math.sqrt(scale_1) * 1./scale_0, 0., 0.
                ],
                [
                    0., 1./ Math.sqrt(scale_1), 0.
                ],
                [
                    0., 0., 1./scale_2
                ]
            ];
            return matrixMultiply(matrixMultiply(matrix1, matrix), matrix2)
        }

        function multiplyMatrixVector(matrix, vector) {
            var result = [];
            for (var i = 0; i < matrix.length; i++) {
                var row = matrix[i];
                var sum = 0;
                for (var j = 0; j < row.length; j++) {
                    sum += row[j] * vector[j];
                }
                result.push(sum);
            }
            return result;
        }

        function addVectors(vec1, vec2) {
            var result = [];
            for (var i = 0; i < vec1.length; i++) {
                result.push(vec1[i] + vec2[i]);
            }
            return result;
        }

        function substractVectors(vec1, vec2) {
            var result = [];
            for (var i = 0; i < vec1.length; i++) {
                result.push(vec1[i] - vec2[i]);
            }
            return result;
        }
    """

    code_part1_1 = (
        code_part1_0
        + """
        function argsort(array) {
            // Create an array of indices [0, 1, 2, ..., array.length - 1]
            var indices = Array.from(array.keys());

            // Sort the indices based on the corresponding values in the array
            indices.sort(function(a, b) {
                return array[a] - array[b];
            });

            return indices;
        }

        const data_rotmatrix_etc = source_rotmatrix_etc.data;

        const col0 = data_rotmatrix_etc['0'];
        const col1 = data_rotmatrix_etc['1'];
        const col2 = data_rotmatrix_etc['2'];
        const rdelta = data_rotmatrix_etc['delta'];
        const rc = data_rotmatrix_etc['center'];
        const width = data_rotmatrix_etc[
            'width'][0] - data_rotmatrix_etc['legend_width'][0];
        const height = data_rotmatrix_etc['height'][0];
        const max_norm = data_rotmatrix_etc['max_norm'][0];
        const xr = data_rotmatrix_etc['xrange'];
        const yr = data_rotmatrix_etc['yrange'];
        const size_coef = data_rotmatrix_etc['size_coef'];
        const rmatrix = [col0, col1, col2];
        const data = source.data;
    """
    )

    code_part1_2 = (
        """
        const excludedKeys = ['x', 'y', 'z', 'color', 'x_saved', 'y_saved', 'z_saved',
                              'point_size', 'point_size_ref',
                              'line_color', 'line_width'];
        const other_keys = Object.keys(data).filter(key => !excludedKeys.includes(key));
        const x = data['x'];
        const y = data['y'];
        const z = data['z'];
        const color = data['color'];
        var x_tmp = Array(x.length).fill(0);
        var y_tmp = Array(x.length).fill(0);
        var z_tmp = Array(x.length).fill(0);
        var color_tmp = Array(x.length).fill(0);
        var label_tmp = Array(x.length).fill(0);
        var x_saved_tmp = Array(x.length).fill(0);
        var y_saved_tmp = Array(x.length).fill(0);
        var z_saved_tmp = Array(x.length).fill(0);
        const x_saved = data['x_saved'];
        const y_saved = data['y_saved'];
        const z_saved = data['z_saved'];
        const point_s = data['point_size'];
        const point_s_ref = data['point_size_ref'];
        const hA = data['highlighted_A'];
        const hB = data['highlighted_B'];
        const line_c = data['line_color'];
        const line_w = data['line_width'];
        var pst = Array(x.length).fill(0);
        var other_keys_array_tmp = Array.from(
            { length: other_keys.length }, () => Array(x.length).fill(0));
        const scaleCoef = (
            yr[1]-yr[0] + 0.00000001)/(xr[1]-xr[0] + 0.00000001) * width/height;
        const rotmataxis = scaleTransform(
            rotationMatrix(rotaxis, f * Math.PI/180.0), 1., scaleCoef, 1.);
        const rot_m = matrixMultiply(rotmataxis, rmatrix);
        const delta_vec = addVectors(
            rc, multiplyMatrixVector(rotmataxis, substractVectors(rdelta, rc)));
        var rot_result = [0, 0, 0];
        """
        + (
            """
        const dim1 = 'original_embedding_' + (slc1.active - 1).toString();
        const dimif1 = (slc1.active > 0);
        const dim2 = 'original_embedding_' + (slc2.active - 1).toString();
        const dimif2 = (slc2.active > 0);
        const dim3 = 'original_embedding_' + (slc3.active - 1).toString();
        const dimif3 = (slc3.active > 0);
        """
            if higher_dim
            else ""
        )
        + """
        for (let i = 0; i < x.length; i++) {
            """
        + (
            """
            x_saved[i] = 0.;
            y_saved[i] = 0.;
            z_saved[i] = 0.;
            if (dimif1) {
                x_saved[i] += data[dim1][i];
            }
            if (dimif2) {
                y_saved[i] += data[dim2][i];
            }
            if (dimif3) {
                z_saved[i] += data[dim3][i];
            }
            """
            if higher_dim
            else ""
        )
        + """
            rot_result = multiplyMatrixVector(
                rot_m, [x_saved[i], y_saved[i], z_saved[i]]);
            x_tmp[i] = rot_result[0] + delta_vec[0];
            y_tmp[i] = rot_result[1] + delta_vec[1];
            z_tmp[i] = rot_result[2] + delta_vec[2];
            color_tmp[i] = color[i];
            x_saved_tmp[i] = x_saved[i];
            y_saved_tmp[i] = y_saved[i];
            z_saved_tmp[i] = z_saved[i];
            pst[i] = 5. / (1. + Math.exp(Math.min(Math.max(
                (-z_tmp[i] + rc[2])/ (max_norm + 0.0000001), -1.), 1.)));
            for (let j = 0; j < other_keys.length; j++) {
                other_keys_array_tmp[j][i] = data[other_keys[j]][i]
            }
        }
        var new_order = argsort(z_tmp);
        var transform_indices = Array(x.length).fill(0);
        for (let i = 0; i < x.length; i++) {
            var k = new_order[i];
            transform_indices[k] = i;
            x[i] = x_tmp[k];
            y[i] = y_tmp[k];
            z[i] = z_tmp[k];
            color[i] = color_tmp[k];
            x_saved[i] = x_saved_tmp[k];
            y_saved[i] = y_saved_tmp[k];
            z_saved[i] = z_saved_tmp[k];
            point_s_ref[i] = pst[k];
            for (let j = 0; j < other_keys.length; j++) {
                data[other_keys[j]][i] = other_keys_array_tmp[j][k]
            }"""
        + size_redef_code
        + """
        }

        const indices = source.selected.indices
        for (let i = 0; i < indices.length; i++) {
            indices[i] = transform_indices[indices[i]]
        }

        source_rotmatrix_etc.change.emit();
        source.change.emit();
    """
    )

    code_part2_1 = (
        code_part1_0
        + """
        const data_rotmatrix_etc = source_rotmatrix_etc.data;

        const col0 = data_rotmatrix_etc['0'];
        const col1 = data_rotmatrix_etc['1'];
        const col2 = data_rotmatrix_etc['2'];
        const rdelta = data_rotmatrix_etc['delta'];
        const rc = data_rotmatrix_etc['center'];
        const width = data_rotmatrix_etc[
            'width'][0] - data_rotmatrix_etc['legend_width'][0];
        const height = data_rotmatrix_etc['height'][0];
        const xr = data_rotmatrix_etc['xrange'];
        const yr = data_rotmatrix_etc['yrange'];
        const offset = data_rotmatrix_etc['offset_angle'];
        const rmatrix = [col0, col1, col2];
    """
    )

    code_part2_2 = """
        const scaleCoef = (
            yr[1]-yr[0] + 0.00000001)/(xr[1]-xr[0] + 0.00000001) * width/height;
        const rotmataxis = scaleTransform(
            rotationMatrix(rotaxis, f * Math.PI/180.0), 1., scaleCoef, 1.);
        const rot_m = matrixMultiply(rotmataxis, rmatrix);
        const delta_vec = addVectors(
            rc, multiplyMatrixVector(rotmataxis, substractVectors(rdelta, rc)));

        col0[0] = rot_m[0][0];
        col0[1] = rot_m[0][1];
        col0[2] = rot_m[0][2];
        col1[0] = rot_m[1][0];
        col1[1] = rot_m[1][1];
        col1[2] = rot_m[1][2];
        col2[0] = rot_m[2][0];
        col2[1] = rot_m[2][1];
        col2[2] = rot_m[2][2];

        rdelta[0] = delta_vec[0];
        rdelta[1] = delta_vec[1];
        rdelta[2] = delta_vec[2];

    """

    reset_slider_yaw = (
        """
            if (bsy.value != 0) {
        """
        + code_part2_1
        + """
            const offset_x = offset[0];
            const f = bsy.value - offset_x;
            const rotaxis = [0, 1, 0];
        """
        + code_part2_2
        + """
            offset[0] = 0;
            source_rotmatrix_etc.change.emit();
            bsy.value = 0;
            bsy.change.emit();
            }
    """
    )
    reset_slider_pitch = (
        """
            if (bsp.value != 0) {
        """
        + code_part2_1
        + """
            const offset_y = offset[1];
            const f = bsp.value - offset_y;
            const rotaxis = [1, 0, 0];
        """
        + code_part2_2
        + """
            offset[1] = 0;
            source_rotmatrix_etc.change.emit();
            bsp.value = 0;
            bsp.change.emit();
            }
    """
    )
    reset_slider_roll = (
        """
            if (bsr.value != 0) {
        """
        + code_part2_1
        + """
            const offset_z = offset[2];
            const f = bsr.value - offset_z;
            const rotaxis = [0, 0, 1];
        """
        + code_part2_2
        + """
            offset[2] = 0;
            source_rotmatrix_etc.change.emit();
            bsr.value = 0;
            bsr.change.emit();
            }
    """
    )

    if higher_dim:
        callback_js2 = bokeh.models.CustomJS(
            args=dict(
                bta=bt_toggle_anim,
                slc1=sl_component1,
                slc2=sl_component2,
                slc3=sl_component3,
                bsp=bt_slider_pitch,
                bsr=bt_slider_roll,
                source=source,
                source_rotmatrix_etc=source_rotmatrix_etc,
            ),
            code="""bta.active = false;"""
            + reset_slider_pitch
            + reset_slider_roll
            + code_part1_1
            + """
                const offset_x = data_rotmatrix_etc['offset_angle'][0];
                const f = cb_obj.value - offset_x;
                const rotaxis = [0, 1, 0];
            """
            + code_part1_2,
        )
    else:
        callback_js2 = bokeh.models.CustomJS(
            args=dict(
                bta=bt_toggle_anim,
                bsp=bt_slider_pitch,
                bsr=bt_slider_roll,
                source=source,
                source_rotmatrix_etc=source_rotmatrix_etc,
            ),
            code="""bta.active = false;"""
            + reset_slider_pitch
            + reset_slider_roll
            + code_part1_1
            + """
                const offset_x = data_rotmatrix_etc['offset_angle'][0];
                const f = cb_obj.value - offset_x;
                const rotaxis = [0, 1, 0];
            """
            + code_part1_2,
        )

    callback_js3 = bokeh.models.CustomJS(
        args=dict(source=source, source_rotmatrix_etc=source_rotmatrix_etc),
        code=code_part2_1
        + """
            const offset_x = offset[0];
            const f = cb_obj.value - offset_x;
            const rotaxis = [0, 1, 0];
        """
        + code_part2_2
        + """
            offset[0] = 0;
            source_rotmatrix_etc.change.emit();
            cb_obj.value = 0;
            cb_obj.change.emit();
        """,
    )

    code_full = """
        const new_offset = data_rotmatrix_etc['offset_angle'];
        const new_offset_x = new_offset[0];
        const new_f = cb_obj.value - new_offset_x;
        const new_rotaxis = [0, 1, 0];
        const new_scaleCoef = (
            yr[1]-yr[0] + 0.00000001)/(xr[1]-xr[0] + 0.00000001) * width/height;
        const new_rotmataxis = scaleTransform(
            rotationMatrix(new_rotaxis, new_f * Math.PI/180.0), 1., new_scaleCoef, 1.);
        const new_rot_m = matrixMultiply(new_rotmataxis, rmatrix);
        const new_delta_vec = addVectors(
            rc, multiplyMatrixVector(new_rotmataxis, substractVectors(rdelta, rc)));

        col0[0] = new_rot_m[0][0];
        col0[1] = new_rot_m[0][1];
        col0[2] = new_rot_m[0][2];
        col1[0] = new_rot_m[1][0];
        col1[1] = new_rot_m[1][1];
        col1[2] = new_rot_m[1][2];
        col2[0] = new_rot_m[2][0];
        col2[1] = new_rot_m[2][1];
        col2[2] = new_rot_m[2][2];

        rdelta[0] = new_delta_vec[0];
        rdelta[1] = new_delta_vec[1];
        rdelta[2] = new_delta_vec[2];
        new_offset[0] = 0;
        source_rotmatrix_etc.change.emit();
        cb_obj.value = 0;
        cb_obj.change.emit();
    """

    if higher_dim:
        callback_full = bokeh.models.CustomJS(
            args=dict(
                slc1=sl_component1,
                slc2=sl_component2,
                slc3=sl_component3,
                bsp=bt_slider_pitch,
                bsr=bt_slider_roll,
                source=source,
                source_rotmatrix_etc=source_rotmatrix_etc,
            ),
            code=reset_slider_pitch
            + reset_slider_roll
            + code_part1_1
            + """
                const offset_x = data_rotmatrix_etc['offset_angle'][0];
                const f = cb_obj.value - offset_x;
                const rotaxis = [0, 1, 0];
            """
            + code_part1_2
            + code_full,
        )
    else:
        callback_full = bokeh.models.CustomJS(
            args=dict(
                bsp=bt_slider_pitch,
                bsr=bt_slider_roll,
                source=source,
                source_rotmatrix_etc=source_rotmatrix_etc,
            ),
            code=reset_slider_pitch
            + reset_slider_roll
            + code_part1_1
            + """
                const offset_x = data_rotmatrix_etc['offset_angle'][0];
                const f = cb_obj.value - offset_x;
                const rotaxis = [0, 1, 0];
            """
            + code_part1_2
            + code_full,
        )

    bt_slider_yaw.js_on_change("value", callback_js2)
    bt_slider_yaw.js_on_change("value_throttled", callback_js3)

    bt_hidden_slider_yaw.js_on_change("value", callback_full)

    if higher_dim:
        callback_js4 = bokeh.models.CustomJS(
            args=dict(
                bta=bt_toggle_anim,
                slc1=sl_component1,
                slc2=sl_component2,
                slc3=sl_component3,
                bsr=bt_slider_roll,
                bsy=bt_slider_yaw,
                source=source,
                source_rotmatrix_etc=source_rotmatrix_etc,
            ),
            code="""bta.active = false;"""
            + reset_slider_roll
            + reset_slider_yaw
            + code_part1_1
            + """
                const offset_y = data_rotmatrix_etc['offset_angle'][1];
                const f = cb_obj.value - offset_y;
                const rotaxis = [1, 0, 0];
            """
            + code_part1_2,
        )
    else:
        callback_js4 = bokeh.models.CustomJS(
            args=dict(
                bta=bt_toggle_anim,
                bsr=bt_slider_roll,
                bsy=bt_slider_yaw,
                source=source,
                source_rotmatrix_etc=source_rotmatrix_etc,
            ),
            code="""bta.active = false;"""
            + reset_slider_roll
            + reset_slider_yaw
            + code_part1_1
            + """
                const offset_y = data_rotmatrix_etc['offset_angle'][1];
                const f = cb_obj.value - offset_y;
                const rotaxis = [1, 0, 0];
            """
            + code_part1_2,
        )

    callback_js5 = bokeh.models.CustomJS(
        args=dict(source=source, source_rotmatrix_etc=source_rotmatrix_etc),
        code=code_part2_1
        + """
            const offset_y = offset[1];
            const f = cb_obj.value - offset_y;
            const rotaxis = [1, 0, 0];
        """
        + code_part2_2
        + """
            offset[1] = 0;
            source_rotmatrix_etc.change.emit();
            cb_obj.value = 0;
            cb_obj.change.emit();
        """,
    )

    bt_slider_pitch.js_on_change("value", callback_js4)
    bt_slider_pitch.js_on_change("value_throttled", callback_js5)

    if higher_dim:
        callback_js6 = bokeh.models.CustomJS(
            args=dict(
                bta=bt_toggle_anim,
                slc1=sl_component1,
                slc2=sl_component2,
                slc3=sl_component3,
                bsp=bt_slider_pitch,
                bsy=bt_slider_yaw,
                source=source,
                source_rotmatrix_etc=source_rotmatrix_etc,
            ),
            code="""bta.active = false;"""
            + reset_slider_pitch
            + reset_slider_yaw
            + code_part1_1
            + """
                const offset_z = data_rotmatrix_etc['offset_angle'][2];
                const f = cb_obj.value - offset_z;
                const rotaxis = [0, 0, 1];
            """
            + code_part1_2,
        )
    else:
        callback_js6 = bokeh.models.CustomJS(
            args=dict(
                bta=bt_toggle_anim,
                bsp=bt_slider_pitch,
                bsy=bt_slider_yaw,
                source=source,
                source_rotmatrix_etc=source_rotmatrix_etc,
            ),
            code="""bta.active = false;"""
            + reset_slider_pitch
            + reset_slider_yaw
            + code_part1_1
            + """
                const offset_z = data_rotmatrix_etc['offset_angle'][2];
                const f = cb_obj.value - offset_z;
                const rotaxis = [0, 0, 1];
            """
            + code_part1_2,
        )

    callback_js7 = bokeh.models.CustomJS(
        args=dict(source=source, source_rotmatrix_etc=source_rotmatrix_etc),
        code=code_part2_1
        + """
            const offset_z = offset[2];
            const f = cb_obj.value - offset_z;
            const rotaxis = [0, 0, 1];
        """
        + code_part2_2
        + """
            offset[2] = 0;
            source_rotmatrix_etc.change.emit();
            cb_obj.value = 0;
            cb_obj.change.emit();
        """,
    )

    bt_slider_roll.js_on_change("value", callback_js6)
    bt_slider_roll.js_on_change("value_throttled", callback_js7)

    if higher_dim:
        slc_data = {
            "1": [sl_component1.active],
            "2": [sl_component2.active],
            "3": [sl_component3.active],
        }
        slc_source = bokeh.models.ColumnDataSource(data=slc_data)

        code_callback_slc = (
            code_part1_1
            + """
            const excludedKeys = ['x', 'y', 'z', 'color', 'x_saved', 'y_saved',
                                'z_saved', 'point_size', 'point_size_ref',
                                'line_color', 'line_width'];
            const other_keys = Object.keys(data).filter(
                key => !excludedKeys.includes(key));
            const x = data['x'];
            const y = data['y'];
            const z = data['z'];
            const color = data['color'];
            var x_tmp = Array(x.length).fill(0);
            var y_tmp = Array(x.length).fill(0);
            var z_tmp = Array(x.length).fill(0);
            var color_tmp = Array(x.length).fill(0);
            var label_tmp = Array(x.length).fill(0);
            var x_saved_tmp = Array(x.length).fill(0);
            var y_saved_tmp = Array(x.length).fill(0);
            var z_saved_tmp = Array(x.length).fill(0);
            const x_saved = data['x_saved'];
            const y_saved = data['y_saved'];
            const z_saved = data['z_saved'];
            const point_s = data['point_size'];
            const point_s_ref = data['point_size_ref'];
            const hA = data['highlighted_A'];
            const hB = data['highlighted_B'];
            const line_c = data['line_color'];
            const line_w = data['line_width'];
            var pst = Array(x.length).fill(0);
            var other_keys_array_tmp = Array.from(
                { length: other_keys.length }, () => Array(x.length).fill(0));
            const rot_m = rmatrix;
            const delta_vec = rdelta;
            var rot_result = [0, 0, 0];
            const dim1 = 'original_embedding_' + (cb_obj.data['1'] - 1).toString();
            const dimif1 = (cb_obj.data['1'] > 0);
            const dim2 = 'original_embedding_' + (cb_obj.data['2'] - 1).toString();
            const dimif2 = (cb_obj.data['2'] > 0);
            const dim3 = 'original_embedding_' + (cb_obj.data['3'] - 1).toString();
            const dimif3 = (cb_obj.data['3'] > 0);
            for (let i = 0; i < x.length; i++) {
                x_saved[i] = 0.;
                y_saved[i] = 0.;
                z_saved[i] = 0.;
                if (dimif1) {
                    x_saved[i] += data[dim1][i];
                }
                if (dimif2) {
                    y_saved[i] += data[dim2][i];
                }
                if (dimif3) {
                    z_saved[i] += data[dim3][i];
                }
                rot_result = multiplyMatrixVector(
                    rot_m, [x_saved[i], y_saved[i], z_saved[i]]);
                x_tmp[i] = rot_result[0] + delta_vec[0];
                y_tmp[i] = rot_result[1] + delta_vec[1];
                z_tmp[i] = rot_result[2] + delta_vec[2];
                color_tmp[i] = color[i];
                x_saved_tmp[i] = x_saved[i];
                y_saved_tmp[i] = y_saved[i];
                z_saved_tmp[i] = z_saved[i];
                pst[i] = 5. / (1. + Math.exp(Math.min(Math.max(
                    (-z_tmp[i] + rc[2])/ (max_norm + 0.0000001), -1.), 1.)));
                for (let j = 0; j < other_keys.length; j++) {
                    other_keys_array_tmp[j][i] = data[other_keys[j]][i]
                }
            }
            var new_order = argsort(z_tmp);
            var transform_indices = Array(x.length).fill(0);
            for (let i = 0; i < x.length; i++) {
                var k = new_order[i];
                transform_indices[k] = i;
                x[i] = x_tmp[k];
                y[i] = y_tmp[k];
                z[i] = z_tmp[k];
                color[i] = color_tmp[k];
                x_saved[i] = x_saved_tmp[k];
                y_saved[i] = y_saved_tmp[k];
                z_saved[i] = z_saved_tmp[k];
                point_s_ref[i] = pst[k];
                for (let j = 0; j < other_keys.length; j++) {
                    data[other_keys[j]][i] = other_keys_array_tmp[j][k]
                }"""
            + size_redef_code
            + """
            }
            const indices = source.selected.indices
            for (let i = 0; i < indices.length; i++) {
                indices[i] = transform_indices[indices[i]]
            }
            var sum_x = 0.;
            var sum_y = 0.;
            var sum_z = 0.;
            if (indices.length > 0) {
                for (let i = 0; i < indices.length; i++) {
                    sum_x = sum_x + x[indices[i]];
                    sum_y = sum_y + y[indices[i]];
                    sum_z = sum_z + z[indices[i]];
                }
                rc[0] = sum_x / indices.length;
                rc[1] = sum_y / indices.length;
                rc[2] = sum_z / indices.length;
            } else {
                for (let i = 0; i < x.length; i++) {
                    sum_x = sum_x + x[i];
                    sum_y = sum_y + y[i];
                    sum_z = sum_z + z[i];
                }
                rc[0] = sum_x / x.length;
                rc[1] = sum_y / x.length;
                rc[2] = sum_z / x.length;
            }
            source_rotmatrix_etc.change.emit();
            source.change.emit();
        """
        )

        sl_component1.js_on_change(
            "active",
            bokeh.models.CustomJS(
                args=dict(
                    slcs=slc_source,
                ),
                code="""
                    // Full copy to trigger the change callback
                    var t = {};
                    for (const key in slcs.data) {
                        t[key] = [];
                        t[key].push(slcs.data[key][0]);
                    }
                    t['1'][0] = cb_obj.active;
                    slcs.data = t;
                    slcs.change.emit();
                """,
            ),
        )
        sl_component2.js_on_change(
            "active",
            bokeh.models.CustomJS(
                args=dict(
                    slcs=slc_source,
                ),
                code="""
                    // Full copy to trigger the change callback
                    var t = {};
                    for (const key in slcs.data) {
                        t[key] = [];
                        t[key].push(slcs.data[key][0]);
                    }
                    t['2'][0] = cb_obj.active;
                    slcs.data = t;
                    slcs.change.emit();
                """,
            ),
        )
        sl_component3.js_on_change(
            "active",
            bokeh.models.CustomJS(
                args=dict(
                    slcs=slc_source,
                ),
                code="""
                    // Full copy to trigger the change callback
                    var t = {};
                    for (const key in slcs.data) {
                        t[key] = [];
                        t[key].push(slcs.data[key][0]);
                    }
                    t['3'][0] = cb_obj.active;
                    slcs.data = t;
                    slcs.change.emit();
                """,
            ),
        )

        slc_source.js_on_change(
            "data",
            bokeh.models.CustomJS(
                args=dict(
                    source=source,
                    source_rotmatrix_etc=source_rotmatrix_etc,
                ),
                code=code_callback_slc,
            ),
        )

    resize_width_input = bokeh.models.TextInput(
        value=str(points_bokeh_plot.width),
        title="Width:",
        name="resize_width",
        width=54,
        margin=(5, 0, 5, 0),
    )
    resize_width_input.js_on_change(
        "value",
        bokeh.models.CustomJS(
            args=dict(pbp=points_bokeh_plot, source_rotmatrix_etc=source_rotmatrix_etc),
            code="""
        var parsed_int = parseInt(this.value);
        if (!isNaN(parsed_int)) {
            pbp.width = parsed_int;
            source_rotmatrix_etc.data['width'][0] = Math.max(parsed_int - 41., 1.);
            source_rotmatrix_etc.change.emit();
            pbp.change.emit();
        }
    """,
        ),
    )

    resize_height_input = bokeh.models.TextInput(
        value=str(points_bokeh_plot.height),
        title="Height:",
        name="resize_height",
        width=54,
        margin=(5, 0, 5, 0),
    )
    resize_height_input.js_on_change(
        "value",
        bokeh.models.CustomJS(
            args=dict(
                pbp=points_bokeh_plot,
                source_rotmatrix_etc=source_rotmatrix_etc,
                slider_range=bt_slider_range,
            ),
            code="""
        var parsed_int = parseInt(this.value);
        if (!isNaN(parsed_int)) {
            pbp.height = parsed_int;
            source_rotmatrix_etc.data['height'][0] = Math.max(parsed_int - 31., 1.);
            source_rotmatrix_etc.change.emit();
            pbp.change.emit();
            //slider_range.value = [ 0.0, 1.0 ];
            //slider_range.change.emit();
        }
    """,
        ),
    )

    def modif_slider_range(attr, old, new):
        current_style = bt_slider_range.stylesheets[0].css
        pattern = r"\{height: \d+px;\}"
        new_style = re.sub(
            pattern, "{height: " + str(max(int(new) - 55, 0)) + "px;}", current_style
        )
        bt_slider_range.stylesheets = [InlineStyleSheet(css=new_style)]

    resize_height_input.on_change(
        "value", lambda attr, old, new: modif_slider_range(attr, old, new)
    )

    sample_search_input = bokeh.models.TextInput(
        value="",
        title="Sample search (enter sample name):",
        name="sample_search_input",
        width=650,
    )

    sample_search_input.js_on_change(
        "value",
        bokeh.models.CustomJS(
            args=dict(source=source),
            code="""
        const data = source.data;
        for (let i=0; i<data["name"].length; i++) {
            if (data["name"][i] == this.value) {
                source.selected.indices = new Array(1).fill(i);
            }
        }
    """,
        ),
    )

    def copy_figure(original, title):
        new_fig = bkp.figure(
            title=title,
            x_axis_label=original.xaxis.axis_label,
            y_axis_label=original.yaxis.axis_label,
        )

        new_fig.width = original.width
        new_fig.height = original.height
        new_fig.background_fill_color = original.background_fill_color
        new_fig.border_fill_color = original.border_fill_color
        new_fig.toolbar_location = "left"
        return new_fig

    violin_plot = copy_figure(points_bokeh_plot, "Violin plots")
    violin_plot.visible = False
    heatmap_plot = copy_figure(points_bokeh_plot, "Heatmap")
    heatmap_plot.visible = False

    resize_width_input_bis = bokeh.models.TextInput(
        value=str(points_bokeh_plot.width),
        title="Width:",
        name="resize_width_bis",
        width=54,
        margin=(5, 0, 5, 0),
    )
    resize_width_input_bis.js_on_change(
        "value",
        bokeh.models.CustomJS(
            args=dict(vp=violin_plot, hp=heatmap_plot),
            code="""
        var parsed_int = parseInt(this.value);
        if (!isNaN(parsed_int)) {
            vp.width = parsed_int;
            hp.width = parsed_int;
            vp.change.emit();
            hp.width.emit();
        }
    """,
        ),
    )

    resize_height_input_bis = bokeh.models.TextInput(
        value=str(points_bokeh_plot.height),
        title="Height:",
        name="resize_height_bis",
        width=54,
        margin=(5, 0, 5, 0),
    )
    resize_height_input_bis.js_on_change(
        "value",
        bokeh.models.CustomJS(
            args=dict(vp=violin_plot, hp=heatmap_plot),
            code="""
        var parsed_int = parseInt(this.value);
        if (!isNaN(parsed_int)) {
            vp.height = parsed_int;
            hp.height = parsed_int;
            vp.change.emit();
            hp.change.emit();
        }
    """,
        ),
    )

    if higher_dim:
        return (
            obs_string,
            obs_numerical,
            points_bokeh_plot,
            violin_plot,
            heatmap_plot,
            bt_slider_point_size,
            bt_hidden_slider_yaw,
            bt_slider_range,
            bt_toggle_anim,
            bt_slider_yaw,
            bt_slider_pitch,
            bt_slider_roll,
            resize_width_input,
            resize_height_input,
            resize_width_input_bis,
            resize_height_input_bis,
            source_rotmatrix_etc,
            div_sample_names,
            sample_search_input,
            sl_component1,
            sl_component2,
            sl_component3,
        )
    else:
        return (
            obs_string,
            obs_numerical,
            points_bokeh_plot,
            violin_plot,
            heatmap_plot,
            bt_slider_point_size,
            bt_hidden_slider_yaw,
            bt_slider_range,
            bt_toggle_anim,
            bt_slider_yaw,
            bt_slider_pitch,
            bt_slider_roll,
            resize_width_input,
            resize_height_input,
            resize_width_input_bis,
            resize_height_input_bis,
            source_rotmatrix_etc,
            div_sample_names,
            sample_search_input,
            None,
            None,
            None,
        )



import bokeh.models
import numpy as np
from pathlib import Path
from PIL import ImageFont
import re
from bokeh.models import InlineStyleSheet


def setup_legend(
    pb_plot,
    obs_string,
    obs_numerical,
    source_rotmatrix_etc,
    resize_width_input,
    bt_slider_range,
):
    source = pb_plot.select(dict(name="scatterplot"))[0].data_source

    hidden_text_label_column = bokeh.models.TextInput(
        value="", title="Label column", name="hidden_text_label_column", width=999
    )
    hidden_text_label_column.js_on_change(
        "value",
        bokeh.models.CustomJS(
            args=dict(source=source, obss=obs_string, obsn=obs_numerical),
            code="""
        if (obss.includes(this.value)) {
            const data = source.data;
            function onlyUnique(value, index, array) {
                return array.indexOf(value) === index;
            }
            var unique = data[this.value].filter(onlyUnique);
            unique.sort();
            var l_values = new Array(unique.length).fill(0);
            const step = 1./(Math.max(unique.length - 1, 1)) * 0.999999;
            var l_values_dict = {};
            l_values_dict[unique[0]] = -1.;
            for (let i = 1; i < l_values.length; i++) {
                l_values[i] = l_values[i-1] + step;
                l_values_dict[unique[i]] = l_values[i]-1.;
            }
            for (let i = 0; i < data["color"].length; i++) {
                data["color"][i] = l_values_dict[data[this.value][i]];
            }
            source.change.emit();
        }
        if (obsn.includes(this.value)) {
            const data = source.data;
            var max_val = Math.max(...data[this.value]);
            var min_val = Math.min(...data[this.value]);
            for (let i = 0; i < data["color"].length; i++) {
                data["color"][i] = (data[this.value][i] - min_val) / (
                    max_val - min_val + 0.000001);
            }
            source.change.emit();
        }
    """,
        ),
    )

    hidden_legend_width = bokeh.models.TextInput(
        value="0", title="Legend width", name="hidden_legend_width", width=999
    )
    hidden_legend_width.js_on_change(
        "value",
        bokeh.models.CustomJS(
            args=dict(source_rotmatrix_etc=source_rotmatrix_etc),
            code="""
        var parsed_int = parseInt(this.value);
        if (!isNaN(parsed_int)) {
            source_rotmatrix_etc.data['legend_width'][0] = parsed_int * 1.;
            source_rotmatrix_etc.change.emit();
        }
    """,
        ),
    )

    def redefine_custom_legend(
        bokeh_plot,
        htls,
        htlc,
        hlw,
        obs_col,
        legend_dict,
        rwi,
        obs_s,
        obs_n,
        bt_slider_range,
    ):
        if obs_col in obs_s:
            bokeh_plot.right = []
            htlc.value = obs_col
            if obs_col in legend_dict:
                legend_list = legend_dict[obs_col][0]
                for legend in legend_list:
                    bokeh_plot.add_layout(legend, "right")
                rwi.value = str(
                    int(bokeh_plot.width - float(hlw.value) + legend_dict[obs_col][1])
                )
                hlw.value = str(legend_dict[obs_col][1])
            else:

                def all_values(arr) -> np.ndarray:
                    av = np.array(list(dict.fromkeys(arr)))
                    av.sort()
                    return av

                data = bokeh_plot.select(dict(name="scatterplot"))[0].data_source.data
                list_vals = all_values(data[obs_col])

                if len(list_vals) == 1:
                    l_values = [0.0]
                else:
                    l_values = (
                        np.arange(0, 1.0000001, 1.0 / (len(list_vals) - 1)) * 0.999999
                    )
                glyph = bokeh_plot.select(dict(name="scatterplot"))[0].glyph
                palette = glyph.fill_color["transform"].palette[0:256]
                l_colors = [palette[int(256 * val)] for val in l_values]
                height = 24
                margin = 0
                spacing = 0
                padding = 5
                max_nr = (bokeh_plot.height - 2 * margin - 2 * padding - height) // (
                    height + spacing
                )
                full_length = len(list_vals)
                cuts = list(np.arange(0, full_length, max_nr)) + [full_length]
                list_intervals = [
                    np.arange(cuts[i], cuts[i + 1]) for i in range(len(cuts) - 1)
                ]
                legend_list = []
                legend_width = 0

                scatter_circles = [
                    bokeh_plot.add_glyph(
                        bokeh.models.Scatter(
                            size=0, x=0, y=0, line_width=0, fill_color=l_colors[i]
                        )
                    )
                    for i in range(full_length)
                ]

                iteration = 0
                for itvl in list_intervals:
                    iteration += 1
                    items_list = [
                        (
                            list_vals[i],
                            [scatter_circles[i]],
                        )
                        for i in itvl
                    ]
                    legend = bokeh.models.Legend(
                        items=items_list,
                        label_height=height,
                        glyph_height=height,
                        spacing=spacing,
                        padding=padding,
                        margin=margin,
                    )
                    legend.click_policy = "mute"
                    for i in range(len(legend.items)):
                        cb_js = bokeh.models.CustomJS(
                            args=dict(
                                htls=htls,
                                label=legend.items[i].label.value,
                                renderer=legend.items[i].renderers[0],
                            ),
                            code="""
                                if (!renderer.muted) {
                                    htls.value = label;
                                }
                                else {
                                    htls.value = "";
                                }
                                renderer.muted = false;
                            """,
                        )
                        legend.items[i].renderers[0].js_on_change("change:muted", cb_js)
                    bokeh_plot.add_layout(legend, "right")
                    legend.label_text_font = "Helvetica"
                    if iteration == 1:
                        label_font_size = legend.label_text_font_size
                        # It's a string like '13px' so need to int-ify it:
                        label_font_size = int(label_font_size[:-2])
                        font = ImageFont.truetype(
                            (Path(__file__).parent.parent / "assets" / "helvetica.ttf")
                            .absolute()
                            .as_posix(),
                            label_font_size,
                        )
                    all_label_width = [
                        font.getlength(x.label["value"]) for x in legend.items
                    ]
                    max_label_width = max(all_label_width)
                    legend_width += (
                        2 * (legend.border_line_width)
                        + legend.glyph_width
                        + max_label_width
                    )
                    legend_list.append(legend)
                rwi.value = str(int(bokeh_plot.width - float(hlw.value) + legend_width))
                hlw.value = str(int(legend_width))
                legend_dict[obs_col] = (legend_list, legend_width)
        bt_slider_range.visible = False
        if obs_col in obs_n:
            bokeh_plot.right = []
            htlc.value = obs_col
            if obs_col in legend_dict:
                legend_list = legend_dict[obs_col][0]
                for legend in legend_list:
                    bokeh_plot.add_layout(legend, "right")
                cbar = legend_list[0]
                min_val = cbar.ticker.ticks[0]
                max_val = cbar.ticker.ticks[-1]
                legend_width = legend_dict[obs_col][1]
                rwi.value = str(int(bokeh_plot.width - float(hlw.value) + legend_width))
                hlw.value = str(legend_width)
            else:
                data = bokeh_plot.select(dict(name="scatterplot"))[0].data_source.data
                max_val = data[obs_col].max()
                min_val = data[obs_col].min()
                viridis_colors = list(bokeh.palettes.Viridis256)
                custom_color_mapper = bokeh.models.LinearColorMapper(
                    palette=viridis_colors, low=min_val, high=max_val
                )
                ltick_vals = [
                    min_val,
                    (max_val + 3.0 * min_val) / 4.0,
                    (max_val + min_val) / 2.0,
                    (3.0 * max_val + min_val) / 4.0,
                    max_val,
                ]
                cbar = bokeh.models.ColorBar(
                    color_mapper=custom_color_mapper,
                    label_standoff=12,
                    width=47,
                    ticker=bokeh.models.FixedTicker(ticks=ltick_vals),
                )
                cbar.major_label_overrides = {
                    nbr: f"""{float(f"{nbr:.3E}"):.10f}""".rstrip("0") + "0"
                    for nbr in ltick_vals
                }
                bokeh_plot.add_layout(cbar, "right")
                tick_strings = list(cbar.major_label_overrides.values())
                label_font_size = cbar.major_label_text_font_size
                label_font_size = int(label_font_size[:-2])
                font = ImageFont.truetype(
                    (Path(__file__).parent.parent / "assets" / "helvetica.ttf")
                    .absolute()
                    .as_posix(),
                    label_font_size,
                )
                all_tick_width = [font.getlength(x) for x in tick_strings]
                max_tick_width = max(all_tick_width)
                legend_width = cbar.width + 23 + max_tick_width
                rwi.value = str(int(bokeh_plot.width - float(hlw.value) + legend_width))
                hlw.value = str(int(legend_width))
                legend_dict[obs_col] = ([cbar], legend_width)
            current_style = bt_slider_range.stylesheets[0].css
            pattern = r"\{margin: 32px 0px 0px -\d+px;\}"
            new_style = re.sub(
                pattern,
                "{margin: 32px 0px 0px -" + str(int(legend_width)) + "px;}",
                current_style,
            )
            bt_slider_range.stylesheets = [InlineStyleSheet(css=new_style)]
            bt_slider_range.start = min_val
            bt_slider_range.end = max_val
            bt_slider_range.value = (min_val, max_val)
            bt_slider_range.step = (max_val - min_val) / 100.0
            bt_slider_range.visible = True

    hidden_text_label_search = bokeh.models.TextInput(
        value="", title="Label search", name="hidden_text_label_search", width=999
    )

    legend_dict = {}

    modifiers_data = {"shift": [0]}

    source_modifiers = bokeh.models.ColumnDataSource(data=modifiers_data)

    pb_plot.js_on_event(
        bokeh.events.Tap,
        bokeh.models.CustomJS(
            args=dict(
                source_modif=source_modifiers,
                htls=hidden_text_label_search,
            ),
            code="""
        const smd = source_modif.data;
        if (cb_obj.modifiers.shift) {
            smd["shift"][0] = 1;
        } else {
            smd["shift"][0] = 0;
        }
        source_modif.change.emit();
    """,
        ),
    )

    hidden_text_label_search.js_on_change(
        "value",
        bokeh.models.CustomJS(
            args=dict(
                source=source,
                source_modif=source_modifiers,
                htlc=hidden_text_label_column,
                hlw=hidden_legend_width,
            ),
            code="""
        const smd = source_modif.data;
        const data = source.data;
        const labels = data[htlc.value];
        if (smd["shift"][0] == 1) {
            if (this.value.slice(-25, this.value.length) != "[-.-.-.-.-shift-.-.-.-.-]")
            {
                const val = this.value;
                this.value = val + "[-.-.-.-.-shift-.-.-.-.-]";
                const indices = source.selected.indices;
                for (let i = 0; i < indices.length; i++) {
                    if (labels[indices[i]] == val) {
                        indices.splice(i, 1);
                    }
                }
                for (let i = 0; i < labels.length; i++) {
                    if (labels[i] == val) {
                        indices.push(i);
                    }
                }
                source.change.emit();
            }
        } else {
            if (this.value.slice(-25, this.value.length) != "[-.-.-.-.-shift-.-.-.-.-]")
            {
                source.selected.indices = [];
                for (let i = 0; i < labels.length; i++) {
                    if (labels[i] == this.value) {
                        source.selected.indices.push(i);
                    }
                }
                source.change.emit();
            }
        }
    """,
        ),
    )

    # menu = [(o_c, o_c) for o_c in obs_string + obs_numerical]
    menu = obs_string + obs_numerical
    select_color_by = bokeh.models.Select(
        title="Color by (select field):", value="", options=menu, width=235
    )
    tooltip = bokeh.models.Tooltip(
        content="Ignores categorical fields with many different unique items (more "
        "than 70 or more than half the number of samples).\u00A0\u00A0",
        position="right",
    )
    help_button = bokeh.models.HelpButton(tooltip=tooltip, margin=(21, 0, 3, 0))

    select_color_by.on_change(
        "value",
        lambda attr, old, new: redefine_custom_legend(
            pb_plot,
            hidden_text_label_search,
            hidden_text_label_column,
            hidden_legend_width,
            new,
            legend_dict,
            resize_width_input,
            obs_string,
            obs_numerical,
            bt_slider_range,
        ),
    )

    return select_color_by, help_button, hidden_text_label_column, hidden_legend_width



# Install with 'pip install -e .'

from setuptools import setup, find_namespace_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="yomix",
    version=(Path(__file__).with_name("yomix") / "VERSION").read_text().strip(),
    description="yomix: an interactive tool to explore low dimensional "
    "embeddings of omics data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/perrin-isir/yomix",
    packages=find_namespace_packages(),
    include_package_data=True,
    license="LICENSE",
)



import bokeh.models


def gene_query_button(offset_text_feature_color):
    bt_open_link = bokeh.models.Button(
        label="← Search these features in HGNC", width=235, height=60
    )

    bt_open_link.js_on_click(
        bokeh.models.CustomJS(
            args=dict(otfc=offset_text_feature_color),
            code="""
            var query_string = otfc.value;
            const ensembl_re = RegExp(/ENS(G|T|P)0[0-9]+/g);
            var q_list = query_string.split('  -  ').join('  +  ').split('  +  '
                            ).filter(element => element);
            q_list = q_list.map(
                elt => [].concat(elt.match(ensembl_re)).filter(
                    elt => elt).concat(elt)).map(elt => elt[0]);
            for (let i=0; i<q_list.length; i++) {
                window.open(
                    "https://www.genenames.org/tools/search/#!/?query=" + q_list[i],
                    '_blank');
            }
        """,
        )
    )

    return bt_open_link
